"""Internal ZincBase class for negative training examples"""

from zincbase.utils.string_utils import split_on

class Negative:
    def __init__(self, expr):
        sub_exprs = split_on(expr, '(', all=False)
        if len(sub_exprs) != 2:
            raise Exception('Syntax error')
        atoms = split_on(sub_exprs[1][:-1], ',')
        self.head = atoms[0]
        self.pred = sub_exprs[0]
        self.tail = atoms[1]
    def __repr__(self):
        return self.pred + '(' + self.head + ', ' + self.tail + ')'



import context

from zincbase import KB
kb = KB()

kb.store('connected(node1, node2)')

node1 = kb.node('node1')

was_called = False
def hello_neighbor(new_neighbor):
    global was_called
    was_called = True
    assert new_neighbor == 'node3'

node1.watch_for_new_neighbor(hello_neighbor)

kb.store('connected(node1, node3)')
assert was_called

node1.grains = 0
def watch_fn(node, prev_val):
    for n, predicate in node.neighbors:
        kb.node(n).grains += 1

node1_watch = node1.watch('grains', watch_fn)
node1 = kb.node('node1')
assert 'grains' in node1._watches

kb.store('connected(node3, node4)')
node2, node3, node4 = kb.node('node2'), kb.node('node3'), kb.node('node4')
node2.grains = 0
node3.grains = 0
node4.grains = 0
node2_watch = node2.watch('grains', watch_fn)
node3_watch = node3.watch('grains', watch_fn)

node1.grains += 1
assert node2.grains == 1
assert node3.grains == 1
assert node4.grains == 1
node3.grains += 1
assert node2.grains == 1
assert node3.grains == 2
assert node4.grains == 2

node4.watch('grains', watch_fn)
kb.store('connected(node4, node5)', node_attributes=[{}, {'grains': 0}])
node5 = kb.node('node5')
node4.grains += 1
assert node4.grains == 3
assert node5.grains == 1

with kb.dont_propagate():
    assert kb._dont_propagate
    node4.grains += 1

assert kb._dont_propagate == False
assert node4.grains == 4
assert node5.grains == 1

assert kb._MAX_RECURSION == 1
times_called = 0
def cycle_watch_fn(node, prev_value):
    global times_called
    times_called += 1
    for n, edges in node.neighbors:
        kb.node(n).grains += 1
kb.store('connected(node_a, node_a)')
node_a = kb.node('node_a')
node_a.grains = 0
node_a.watch('grains', cycle_watch_fn)
node_a.grains = 1
assert times_called == 2
assert node_a.grains == 2

node_a.grains = 3
assert times_called == 4
assert node_a.grains == 4

kb.set_recursion_limit(2)
node_a.grains = 5
assert times_called == 7
assert node_a.grains == 7

kb = KB()
kb.store('node(node1)', node_attributes=[{'value': 0}])
kb.store('node(node2)', node_attributes=[{'value': 0}])
kb.store('node(node3)', node_attributes=[{'value': 0}])
kb.store('connected(node1, node2)')
kb.store('connected(node2, node3)')
def watch_fn(node, prev_val):
    for n, pred in node.neighbors:
        kb.node(n).value += 1
kb.node('node1').watch('value', watch_fn)
kb.node('node2').watch('value', watch_fn)
kb.node('node1').value += 1
assert kb.node('node1').value == 1
assert kb.node('node2').value == 1
assert kb.node('node3').value == 1
kb.set_propagation_limit(1)
kb.node('node1').value += 1
assert kb.node('node1').value == 2
kb.node('node2').value == 2
kb.node('node3').value == 1

kb = KB()
kb.store('is_linked(a,b)', edge_attributes={'zig': 1})
edge = kb.edge('a', 'is_linked', 'b')
node_a = kb.node('a')
node_b = kb.node('b')
node_a.node_zig = 1
node_b.node_zig = 2
assert edge.zig == 1
was_called = False
def nights_watch(edge, prev_val):
    global was_called
    assert prev_val == 1
    assert edge.nodes[0] == 'a'
    assert edge.nodes[1] == 'b'
    for node in edge.nodes:
        node.node_zig += 1
    was_called = True
edge.watch('zig', nights_watch)
edge.zig += 1
assert edge.zig == 2
assert node_a.node_zig == 2
assert node_b.node_zig == 3
assert was_called

print('All propagation tests passed.')


import csv
import re

from tqdm import tqdm

from zincbase.utils.string_utils import cleanse

def calc_mrr(kb, test_file, delimiter=',', header=None, size=None):
    """Calculate the mean reciprocal rank using a test set."""
    with open(test_file) as f:
        reader = csv.reader(f, delimiter=delimiter)
        if header:
            next(reader, None)
        i = 0
        test_triples = []
        for row in reader:
            pred = cleanse(pred)
            sub = cleanse(sub)
            ob = cleanse(ob)
            if not (sub.replace('_','').isalnum() and ob.replace('_','').isalnum()):
                continue
            test_triples.append((sub, pred, ob))
    mrr = 0
    print('Test set size: {}'.format(len(test_triples)))
    if not size:
        size = len(test_triples)
    for t in tqdm(test_triples[:size]):
        # TODO, this is inefficient, not batched.
        try:
            ob = t[2]
            ranks = kb.get_most_likely(t[0], t[1], '?', k=100)
            ranked_ents = [x['triple'][2] for x in ranks]
            if ob not in ranked_ents:
                continue
            mrr += 1 / (ranked_ents.index(ob) + 1)
        except:
            continue
    return mrr / size



import numpy as np
import torch

from torch.utils.data import Dataset

class NegDataset(Dataset):
    """Zincbase sets this up automatically from the knowledge base.
    It's a generator used for negative examples.
    """
    def __init__(self, neg_triples):
        self.triples = neg_triples
        self.len = len(neg_triples)
    def __len__(self):
        return self.len
    def __getitem__(self, idx):
        t = self.triples[idx]
        return torch.LongTensor(t), torch.LongTensor([[0., 0., 0.]]), torch.FloatTensor([0.]), 'neg', False

class TrainDataset(Dataset):
    """Zincbase sets this up automatically from the knowledge base.
    It's the generator for the RotatE algorithm.
    """

    def __init__(self, triples, nrelation, negative_sample_size, mode):
        self.len = len(triples)
        self.triples = triples
        self.nrelation = nrelation
        self.negative_sample_size = negative_sample_size
        self.mode = mode
        self.count = self.count_frequency(triples)
        self.true_head, self.true_tail = self.get_true_head_and_tail(self.triples)
        self.true_attr = self.get_true_attr(self.triples)
        self.nentity = len(self.true_attr)

    def __len__(self):
        return self.len

    def __getitem__(self, idx):
        positive_sample = self.triples[idx]

        head, relation, tail, attr, true = positive_sample

        subsampling_weight = self.count[(head, relation)] + self.count[(tail, -relation - 1)]
        subsampling_weight = torch.sqrt(1 / torch.Tensor([subsampling_weight]))

        negative_sample_list = []
        negative_sample_size = 0
        # This is not really 'negative samples' but rather a list of head/tails that dont appear
        # with the other part of the tuple
        while negative_sample_size < self.negative_sample_size:
            negative_sample = np.random.randint(self.nentity, size=self.negative_sample_size*2)
            if self.mode == 'head-batch':
                mask = np.in1d(
                    negative_sample, 
                    self.true_head[(relation, tail)], 
                    assume_unique=True, 
                    invert=True
                )
            elif self.mode == 'tail-batch':
                mask = np.in1d(
                    negative_sample, 
                    self.true_tail[(head, relation)], 
                    assume_unique=True, 
                    invert=True
                )
            else:
                raise ValueError('Training batch mode %s not supported' % self.mode)
            negative_sample = negative_sample[mask]
            negative_sample_list.append(negative_sample)
            negative_sample_size += negative_sample.size
        
        negative_sample = np.concatenate(negative_sample_list)[:self.negative_sample_size]
        negative_sample = torch.from_numpy(negative_sample).float()

        tmp = [positive_sample[0], positive_sample[1], positive_sample[2]]
        for item in positive_sample[3:]:
            if isinstance(item, list) or isinstance(item, tuple):
                for subitem in item:
                    tmp.append(subitem)
            else:
                tmp.append(item)
        positive_sample = tmp
        positive_sample = torch.LongTensor(positive_sample) #TODO First 3 needs to be a longtensor, after that should be floats
        return positive_sample, negative_sample, subsampling_weight, self.mode, true

    @staticmethod
    def collate_fn(data):
        positive_sample = torch.stack([_[0] for _ in data], dim=0)
        negative_sample = torch.stack([_[1] for _ in data], dim=0)
        subsample_weight = torch.cat([_[2] for _ in data], dim=0)
        mode = data[0][3]
        true = data[0][4]
        return positive_sample, negative_sample, subsample_weight, mode, true

    @staticmethod
    def count_frequency(triples, start=4):
        count = {}
        for head, relation, tail, attr, true in triples:
            if (head, relation) not in count:
                count[(head, relation)] = start
            else:
                count[(head, relation)] += 1

            if (tail, -relation-1) not in count:
                count[(tail, -relation-1)] = start
            else:
                count[(tail, -relation-1)] += 1
        return count

    @staticmethod
    def get_true_attr(triples):
        true_attr = {}
        for head, relation, tail, attr, true in triples:
            true_attr[head] = attr
        return true_attr

    @staticmethod
    def get_true_head_and_tail(triples):

        true_head = {}
        true_tail = {}

        for head, relation, tail, attr, true in triples:
            if (head, relation) not in true_tail:
                true_tail[(head, relation)] = []
            true_tail[(head, relation)].append(tail)
            if (relation, tail) not in true_head:
                true_head[(relation, tail)] = []
            true_head[(relation, tail)].append(head)

        for relation, tail in true_head:
            true_head[(relation, tail)] = np.array(list(set(true_head[(relation, tail)])))
        for head, relation in true_tail:
            true_tail[(head, relation)] = np.array(list(set(true_tail[(head, relation)])))

        return true_head, true_tail

class BidirectionalOneShotIterator(object):
    """ZincBase uses this class automatically when you want to train a model from a KB.
    """
    def __init__(self, dataloader_head, dataloader_tail, dataloader_neg=None, neg_ratio=1):
        self.iterator_head = self.one_shot_iterator(dataloader_head)
        self.iterator_tail = self.one_shot_iterator(dataloader_tail)
        self.neg = False
        if dataloader_neg:
            self.neg = True
            self.neg_ratio = neg_ratio
            self.iterator_neg = self.one_shot_iterator(dataloader_neg)
        self.step = 0

    def __next__(self):
        if self.neg:
            return self.next_with_neg()
        return self.next_no_neg()

    def next_with_neg(self):
        self.step += 1
        if self.step % self.neg_ratio == 0:
            data = next(self.iterator_neg)
        elif self.step % 2 == 0:
            data = next(self.iterator_head)
        else:
            data = next(self.iterator_tail)
        return data

    def next_no_neg(self):
        self.step += 1
        if self.step % 2 == 0:
            data = next(self.iterator_head)
        else:
            data = next(self.iterator_tail)
        return data

    @staticmethod
    def one_shot_iterator(dataloader):
        while True:
            for data in dataloader:
                yield data



import context

from zincbase import KB

kb = KB()

kb.store('append([ ], List, List)')
kb.store('append([Head | Tail], List, [Head | Result]) :- append(Tail, List, Result)')
b = kb.query('append([a, b], [c, d], X)'); b = list(b); assert len(b) == 1; assert b[0]['X'] == '[a,b,c,d]'
b = kb.query('append([a, b], X, [a, b, c, d])'); b = list(b); assert len(b) == 1; assert b[0]['X'] == '[c,d]'
b = kb.query('append(X, Y, [a, b, c, d])'); b = list(b); assert len(b) == 5; assert b[0]['X'] == '[]'; assert b[0]['Y'] == '[a,b,c,d]'
assert b[1]['X'] == '[a]'; assert b[1]['Y'] == '[b,c,d]'; assert b[2]['X'] == '[a,b]'; assert b[2]['Y'] == '[c,d]'
assert b[3]['X'] == '[a,b,c]'; assert b[3]['Y'] == '[d]'; assert b[4]['X'] == '[a,b,c,d]'; assert b[4]['Y'] == '[]'

print('All list tests passed.')


# Neural Network tests

# Possible that tests could flake given stochasticity of NN
# but they are fairly relaxed.

import context
from zincbase import KB

kb = KB()
kb.seed(555)

people = ['john', 'oleg', 'tom', 'vedant', 'akshay', 'todd']
for person in people:
    kb.store('works_at({}, primer)'.format(person))
other_people = ['other1', 'other2', 'other3', 'other4', 'other5', 'other6']
for person in other_people:
    kb.store('works_at({}, zillow)'.format(person))
kb.store('based_in(primer, bay_area)')
kb.store('based_in(zillow, seattle)')

for person in people:
    kb.store('lives_in({}, bay_area)'.format(person))

for person in other_people:
    kb.store('lives_in({}, seattle)'.format(person))

kb.store('knows(tom, john)'); kb.store('knows(tom, oleg)'); kb.store('knows(akshay, john)'); kb.store('knows(tom, todd)'); kb.store('knows(vedant, akshay)')
kb.store('knows(other2, other1)'); kb.store('knows(other6, other5)'); kb.store('knows(other1, other2)'); kb.store('knows(other4, other3)'); kb.store('knows(other3, other2)')
kb.store('knows(tom, other4)')
kb.store('lives_in(shamala, bay_area)'); kb.store('lives_in(mary, seattle)')
kb.store('associated_with(zillow, amazon)'); kb.store('associated_with(primer, google)')
kb.store('associated_with(amazon, zillow)'); kb.store('associated_with(google, primer)')
kb.store('associated_with(amazon, microsoft)'); kb.store('associated_with(google, facebook)')
kb.store('based_in(microsoft, seattle)'); kb.store('based_in(facebook, bay_area)')

# # # # # # # # # # # # # # # # # # # # # # # #
# Test using node attributes in NN
# # # # # # # # # # # # # # # # # # # # # # # #
for node in ['tom', 'todd', 'oleg', 'john', 'akshay', 'vedant']:
    node = kb.node(node)
    node['owns_a_raincoat'] = 0.0
    node['doesnt_own_raincoat'] = 1.0
for node in ['other1', 'other2', 'other3', 'other4', 'other5', 'other6']:
    node = kb.node(node)
    node['owns_a_raincoat'] = 1.0
    node['doesnt_own_raincoat'] = 0.0
    
kb.build_kg_model(cuda=False, embedding_size=30, node_attributes=['owns_a_raincoat', 'doesnt_own_raincoat'],
               attr_loss_to_graph_loss=0.9)
# Ideally use bs=1 to overfit on this small dataset
# bs=2 at least checks that it works with > 1 bs
kb.train_kg_model(steps=12001, batch_size=2, neg_to_pos=4)

# # # # # # # # # # # # # # # # # # # # # # # #
# People from Seattle should be more likely to
# own an umbrella (attribute prediction test)
# # # # # # # # # # # # # # # # # # # # # # # #
x = kb._kg_model.run_embedding(kb.get_embedding('other1'), 'owns_a_raincoat')
y = kb._kg_model.run_embedding(kb.get_embedding('other1'), 'doesnt_own_raincoat')
assert round(x) == 1; assert round(y) == 0
x = kb._kg_model.run_embedding(kb.get_embedding('other2'), 'owns_a_raincoat')
y = kb._kg_model.run_embedding(kb.get_embedding('other2'), 'doesnt_own_raincoat')
assert round(x) == 1; assert round(y) == 0
x = kb._kg_model.run_embedding(kb.get_embedding('mary'), 'owns_a_raincoat')
y = kb._kg_model.run_embedding(kb.get_embedding('mary'), 'doesnt_own_raincoat')
assert x > 1.5 * y # Not a known fact
x = kb._kg_model.run_embedding(kb.get_embedding('tom'), 'owns_a_raincoat')
y = kb._kg_model.run_embedding(kb.get_embedding('tom'), 'doesnt_own_raincoat')
assert round(x) == 0; assert round(y) == 1
x = kb._kg_model.run_embedding(kb.get_embedding('todd'), 'owns_a_raincoat')
y = kb._kg_model.run_embedding(kb.get_embedding('todd'), 'doesnt_own_raincoat')
assert round(x) == 0; assert round(y) == 1
x = kb._kg_model.run_embedding(kb.get_embedding('shamala'), 'owns_a_raincoat')
y = kb._kg_model.run_embedding(kb.get_embedding('shamala'), 'doesnt_own_raincoat')
assert y > 1.5 * x # Not a known fact

# # # # # # # # # # # # # # # # # # # # # # # #
# These relations should still work (link
# prediction test)
# # # # # # # # # # # # # # # # # # # # # # # #
bay_prob = kb.estimate_triple_prob('tom', 'lives_in', 'bay_area')
sea_prob = kb.estimate_triple_prob('tom', 'lives_in', 'seattle')
assert bay_prob > 2 * sea_prob

bay_prob = kb.estimate_triple_prob('shamala', 'lives_in', 'bay_area')
sea_prob = kb.estimate_triple_prob('shamala', 'lives_in', 'seattle')
assert bay_prob > 2 * sea_prob

sea_prob = kb.estimate_triple_prob('other1', 'lives_in', 'seattle')
bay_prob = kb.estimate_triple_prob('other1', 'lives_in', 'bay_area')
assert sea_prob > 2 * bay_prob
sea_prob = kb.estimate_triple_prob('other2', 'lives_in', 'seattle')
bay_prob = kb.estimate_triple_prob('other2', 'lives_in', 'bay_area')
assert sea_prob > 2 * bay_prob

print('First suite of neural network tests passed.')
# # # # # # # # # # # # # # # # # # # # # # # #
# Test predicate attributes
# # # # # # # # # # # # # # # # # # # # # # # #
kb.store('lives_in(tom, seattle)') ## TODO seems weird to have to add this fact then stick 'formerly' on it
# on the next line.
kb.edge('tom', 'lives_in', 'seattle')['formerly'] = 1.0
kb.seed(555)
kb.build_kg_model(cuda=False, embedding_size=50, node_attributes=['owns_a_raincoat', 'doesnt_own_raincoat'],
                pred_attributes=['formerly'], attr_loss_to_graph_loss=0.9, pred_loss_to_graph_loss=5.0)
kb.train_kg_model(steps=8001, batch_size=2, neg_to_pos=4)

# # # # # # # # # # # # # # # # # # # # # # # #
# People from Seattle should be more likely to
# own an umbrella (attribute prediction test)
# # # # # # # # # # # # # # # # # # # # # # # #
x = kb._kg_model.run_embedding(kb.get_embedding('other1'), 'owns_a_raincoat')
y = kb._kg_model.run_embedding(kb.get_embedding('other1'), 'doesnt_own_raincoat')
assert round(x) == 1; assert round(y) == 0
x = kb._kg_model.run_embedding(kb.get_embedding('other2'), 'owns_a_raincoat')
y = kb._kg_model.run_embedding(kb.get_embedding('other2'), 'doesnt_own_raincoat')
assert round(x) == 1; assert round(y) == 0
x = kb._kg_model.run_embedding(kb.get_embedding('mary'), 'owns_a_raincoat')
y = kb._kg_model.run_embedding(kb.get_embedding('mary'), 'doesnt_own_raincoat')
assert x > 1.5 * y # Not a known fact
x = kb._kg_model.run_embedding(kb.get_embedding('tom'), 'owns_a_raincoat')
y = kb._kg_model.run_embedding(kb.get_embedding('tom'), 'doesnt_own_raincoat')
assert round(x) == 0; assert round(y) == 1
x = kb._kg_model.run_embedding(kb.get_embedding('todd'), 'owns_a_raincoat')
y = kb._kg_model.run_embedding(kb.get_embedding('todd'), 'doesnt_own_raincoat')
assert round(x) == 0; assert round(y) == 1
x = kb._kg_model.run_embedding(kb.get_embedding('shamala'), 'owns_a_raincoat')
y = kb._kg_model.run_embedding(kb.get_embedding('shamala'), 'doesnt_own_raincoat')
assert y > 1.5 * x # Not a known fact

# # # # # # # # # # # # # # # # # # # # # # # #
# These relations should still work (link
# prediction test)
# # # # # # # # # # # # # # # # # # # # # # # #
bay_prob = kb.estimate_triple_prob('tom', 'lives_in', 'bay_area')
sea_prob = kb.estimate_triple_prob('tom', 'lives_in', 'seattle')
assert bay_prob > sea_prob

bay_prob = kb.estimate_triple_prob('shamala', 'lives_in', 'bay_area')
sea_prob = kb.estimate_triple_prob('shamala', 'lives_in', 'seattle')
assert bay_prob > 2 * sea_prob

sea_prob = kb.estimate_triple_prob('other1', 'lives_in', 'seattle')
bay_prob = kb.estimate_triple_prob('other1', 'lives_in', 'bay_area')
assert sea_prob > 2 * bay_prob
sea_prob = kb.estimate_triple_prob('other2', 'lives_in', 'seattle')
bay_prob = kb.estimate_triple_prob('other2', 'lives_in', 'bay_area')
assert sea_prob > 2 * bay_prob

x = kb.estimate_triple_prob_with_attrs('tom', 'lives_in', 'bay_area', 'formerly')
y = kb.estimate_triple_prob_with_attrs('john', 'lives_in', 'bay_area', 'formerly')
z = kb.estimate_triple_prob_with_attrs('mary', 'lives_in', 'bay_area', 'formerly')
assert x > y; assert x > z

x = kb.estimate_triple_prob_with_attrs('tom', 'knows', 'john', 'formerly')
y = kb.estimate_triple_prob('tom', 'knows', 'john')
assert y > x

print('Neural network tests passed.')


from zincbase.utils.string_utils import split_on

from zincbase.logic.Term import Term

class Rule:
    def __init__(self, expr, kb=None):
        parts = split_on(expr, ':-')
        self.head = Term(parts[0], kb=kb)
        self.goals = []
        if len(parts) == 2:
            sub_goals = split_on(parts[1], ',')
            for sub_goal in sub_goals:
                self.goals.append(Term(sub_goal, kb=kb))
    def __repr__(self):
        return str(self.head)


from collections import defaultdict
import copy

import networkx as nx

class Edge:
    """Class representing an edge in the KB.
    """
    def __init__(self, kb, sub, pred, ob, data={}, watches=[]):
        super().__setattr__('_kb', kb)
        super().__setattr__('_name', str(sub) + '___' + str(pred) + '___' + str(ob))
        super().__setattr__('_sub', str(sub))
        super().__setattr__('_pred', str(pred))
        super().__setattr__('_ob', str(ob))
        super().__setattr__('_recursion_depth', 0)
        super().__setattr__('_watches', defaultdict(list))
        super().__setattr__('_edge', self._kb.G[self._sub][self._ob])
        for watch in watches:
            self._watches[watch[0]].append(watch[1])
    
    def __repr__(self):
        return self._name
    
    def __eq__(self, comparator):
        return self._name == str(comparator)
    
    def __ne__(self, comparator):
        return self._name != str(comparator)

    def __iter__(self):
        for attr in self.attrs:
            yield(attr)

    def __getattr__(self, key):
        try:
            for _, edge in self._edge.items():
                if edge['pred'] == self._pred:
                    return edge[key]
        except KeyError as e:
            return None

    def __setattr__(self, key, value):
        if self._kb._global_propagations > self._kb._PROPAGATION_LIMIT:
            return False
        if self._recursion_depth > self._kb._MAX_RECURSION:
            return False
        self._kb._global_propagations += 1
        super().__setattr__('_recursion_depth', self._recursion_depth + 1)
        for _, attrs in self._edge.items():
            if attrs['pred'] == self._pred:
                prev_val = attrs.get(key, None)
                attrs.update({key: value})
                if not self._kb._dont_propagate:
                    for watch_fn in self._watches.get(key, []):
                        watch_fn(self, prev_val)
                super().__setattr__('_recursion_depth', self._recursion_depth - 1)
                self._kb._global_propagations -= 1

    def __getitem__(self, key):
        return self.__getattr__(key)
    
    def __setitem__(self, key, value):
        return self.__setattr__(key, value)
    
    def __delitem__(self, attr):
        for _, attrs in self._edge.items():
            if attrs['pred'] == self._pred:
                del attrs[attr]
    
    def get(self, attr, default):
        try:
            return self.attrs[attr]
        except:
            return default

    @property
    def nodes(self):
        """Return the nodes that this edge is connected to as tuple of (subject, object)
        """
        return [self._kb.node(self._sub), self._kb.node(self._ob)]

    @property
    def attrs(self):
        """Returns attributes of the edge stored in the KB
        """
        attributes = None
        for _, edge in self._edge.items():
            if edge['pred'] == self._pred:
                attributes = copy.deepcopy(edge)
        if attributes is None:
            return False
        try:
            del attributes['pred']
            del attributes['_watches']
        except:
            pass
        return attributes
    
    def watch(self, attribute, fn):
        """Execute user-defined function when the value of attribute changes.
        Function takes two args: `edge` which has access to all
        its own attributes, and the second
        arg is the previous value of the attribute that changed.

        As cycles are possible in the graph, changes to an edge attribute, that
        change the attributes of the nodes it's connected to, etc,
        may eventually propagate back to change the original edge's attribute again,
        ad infinitum until the stack explodes. To prevent this, in one "update cycle", more
        than `kb._MAX_RECURSION` updates will be rejected.

        :returns int: id of the watch

        :Example:

        >>> from zincbase import KB
        >>> kb = KB()
        >>> kb.store('edge(a,b)')
        0
        >>> edge = kb.edge('a', 'edge', 'b')
        >>> edge.resistance = 3
        >>> print(edge.resistance)
        3
        >>> edge.watch('resistance', lambda x, prev_val: print('resistance changed to ' + str(x.resistance)))
        ('resistance', 0)
        >>> edge.resistance += 1
        resistance changed to 4

        """
        self._watches[attribute].append(fn)
        return (attribute, len(self._watches) - 1)
    
    def remove_watch(self, attribute_or_watch_id):
        """Stop watching `attribute_or_watch_id`.
        If it is a string, delete all watches for that attribute.
        If it is a tuple of (attribute, watch_id): delete that specific watch.
        """
        if isinstance(attribute_or_watch_id, tuple):
            self._watches[attribute_or_watch_id[0]].pop(attribute_or_watch_id[1])
        else:
            self._watches[attribute_or_watch_id] = []


"""Test the combination of edge attributes and negative examples."""

import context

from zincbase import KB

kb = KB()
kb.seed(555)

kb.from_csv('./assets/countries_s1_train.csv', delimiter='\t')

# specifying both a ~ and a truthiness < 0 is probably unnecessary.
kb.store('~locatedin(canada, africa)', {'truthiness': -1.})

kb.build_kg_model(cuda=False, embedding_size=100, pred_attributes=['truthiness'])

kb.train_kg_model(steps=1000, batch_size=4, neg_ratio=0.01)

canada_in_africa = kb.estimate_triple_prob('canada', 'locatedin', 'africa')
canada_in_asia = kb.estimate_triple_prob('canada', 'locatedin', 'asia')
canada_in_america = kb.estimate_triple_prob('canada', 'locatedin', 'northern_america')
assert 2 * canada_in_africa < canada_in_asia
assert canada_in_america > 2 * canada_in_asia

clafrica_truthiness = kb.estimate_triple_prob_with_attrs('canada', 'locatedin', 'africa', 'truthiness')
clasia_truthiness = kb.estimate_triple_prob_with_attrs('canada', 'locatedin', 'asia', 'truthiness')

assert clafrica_truthiness < clasia_truthiness

print('All truthiness tests passed.')


import types

import context
from zincbase import KB

kb = KB()
b = kb.store('c(x)'); assert b == 0
b = kb.query('c(X)'); assert isinstance(b, types.GeneratorType); b = list(b); 
assert len(b) == 1; assert b[0]['X'] == 'x'; assert kb.node(b[0]['X']) == 'x'; assert kb.node(b[0]['X']).attrs == {}
b = kb.store('c(y)'); assert b == 1
b = kb.delete_rule(1); assert b; assert not list(kb.query('c(y)'))
b = kb.store('c(y)'); assert b == 1
b = kb.query('c(X)'); b = list(b); assert len(b) == 2; assert b[0]['X'] in ('x', 'y');
assert b[1]['X'] in ('x', 'y'); assert b[0]['X'] != b[1]['X']; assert kb.node(b[0]['X']).attrs == {}
assert kb.node(b[1]['X']).attrs == {}
b = kb.store('loves(tom, shamala)'); assert 'tom' in kb.G; assert 'shamala' in kb.G
assert kb.G['tom']['shamala'][0]['pred'] == 'loves'
b = kb.query('loves(X, Y)'); b = list(b); assert len(b) == 1; assert b[0]['X'] == 'tom'; assert b[0]['Y'] == 'shamala'
b = kb.query('loves( X,Y)'); b = list(b); assert len(b) == 1; assert b[0]['X'] == 'tom'; assert b[0]['Y'] == 'shamala'
b = kb.query('loves(X, foo)'); b = list(b); assert not b
b = kb.query('loves(tom, shamala)'); b = list(b); assert b
b = kb.query('loves(tom, X)'); b = list(b); assert b; assert len(b) == 1; assert b[0]['X'] == 'shamala'
kb.store('loves(kitty, shamala)')
b = kb.query('loves(Who, shamala)'); b = list(b); assert b; assert len(b) == 2;
assert b[0]['Who'] in ('tom', 'kitty'); assert b[1]['Who'] in ('tom', 'kitty'); assert b[0]['Who'] != b[1]['Who']
kb.store('love_rel(X, Y) :- loves(X, Y)')
b = kb.query('love_rel(X, Y)'); b = list(b); assert len(b) == 2; assert b[0]['Y'] == b[1]['Y'] == 'shamala'
assert b[0]['X'] in ('tom', 'kitty'); assert b[1]['X'] in ('tom', 'kitty'); assert b[0]['X'] != b[1]['X']
kb.store('endless(X):-endless(X)')
b = kb.query('endless(zig)'); b = list(b); assert not b
kb.store('eats(kitty, tuna)')
b = kb.query('eats(K, What_Food)'); b = list(b); assert b; assert len(b) == 1; assert b[0]['What_Food'] == 'tuna'
b = kb.store('is_cat(C) :- eats(C, tuna), love_rel(Y, C)')
b = kb.query('is_cat(C)'); b = list(b); assert not b
kb.store('loves(tom, kitty)')
b = kb.query('is_cat(C)'); b = list(b); assert b; assert len(b) == 1; assert b[0]['C'] == 'kitty'
b = kb.query('is_cat(kitty)'); b = list(b); assert b
b = kb.query('is_cat(tom)'); b = list(b); assert not b
b = kb.query('eats(kitty, _)'); b = next(b); assert b; assert isinstance(b, bool)
b = kb.query('eats(_, tuna)'); b = next(b); assert b; assert isinstance(b, bool)
b = kb.query('eats(_, X)'); b = list(b); assert len(b) == 1; assert len(b[0]) == 1; assert b[0]['X'] == 'tuna'
b = kb.query('eats(Y, X)'); b = list(b); assert len(b) == 1; assert b[0]['X'] == 'tuna'; assert b[0]['Y'] == 'kitty'

kb.store('employed_by(tom, primer)')
kb.store('works_as(tom, engineer)')
kb.store('engineers_at(X, Y) :- employed_by(X, Y), works_as(X, engineer)')
b = kb.query('employed_by(X, Y)'); b = list(b); assert b[0]['X'] == 'tom'; assert b[0]['Y'] == 'primer'
b = kb.query('works_as(tom, X)'); b = list(b); assert b; assert b[0]['X'] == 'engineer'
b = kb.query('engineers_at(ZZ, primer)'); b = list(b); assert len(b) == 1; assert b[0]['ZZ'] == 'tom'

kb.store('employed_by(oleg,primer)')
b = kb.query('engineers_at(ZZZ, primer)'); b = list(b); assert len(b) == 1; assert b[0]['ZZZ'] == 'tom'
kb.store('works_as(oleg, engineer)')
b = kb.query('engineers_at(Xss, primer)'); b = list(b); assert len(b) == 2; assert b[0]['Xss'] in ['oleg', 'tom']; assert b[1]['Xss'] in ['oleg', 'tom']; assert b[0]['Xss'] != b[1]['Xss']

kb.store('engineers_at_primer(You) :- engineers_at(You, primer)')
b = kb.query('engineers_at_primer(Engineers)'); b = list(b); assert b[0]['Engineers'] in ['oleg', 'tom'];
assert b[1]['Engineers'] in ['oleg', 'tom']; assert b[1]['Engineers'] != b[0]['Engineers']

kb.store('X_is_engineer_at_primer_doing_neural_networks(X) :- engineers_at_primer(X), works_on(X, neural_networks)')
b = kb.query('X_is_engineer_at_primer_doing_neural_networks(tom)'); b = list(b); assert not b

kb.store('works_on(tom, neural_networks)')
b = kb.query('X_is_engineer_at_primer_doing_neural_networks(tom)'); b = next(b); assert b
b = kb.query('X_is_engineer_at_primer_doing_neural_networks(_)'); b = next(b); assert b; assert isinstance(b, bool)
b = kb.query('X_is_engineer_at_primer_doing_neural_networks(Brainiac)'); b = list(b); assert b; assert len(b) == 1; assert b[0]['Brainiac'] == 'tom'
kb.store('works_on(oleg, neural_networks)')
b = kb.query('X_is_engineer_at_primer_doing_neural_networks(Brainiac)'); b = list(b); assert b; assert len(b) == 2; assert b[0]['Brainiac'] in ['oleg', 'tom']
assert b[1]['Brainiac'] in ['oleg', 'tom']; assert b[1]['Brainiac'] != b[0]['Brainiac'];

kb = KB()
kb.store('gave_mono_to(john, jane)')
kb.store('gave_mono_to(jane, phil)')
b = kb.bfs('john', 'phil'); b = list(b); assert len(b) == 1 
assert len(b[0]) == 2; assert b[0][0] == ('gave_mono_to', 'jane'); assert b[0][1] == ('gave_mono_to', 'phil')

b = kb.bfs('phil', 'john'); b = list(b); assert not b

b = kb.bfs('phil', 'john', reverse=True); b = list(b); assert len(b) == 1
assert b[0][0] == ('gave_mono_to', 'jane'); assert b[0][1] == ('gave_mono_to', 'john')

kb.store('gave_mono_to(jane, janice)')
kb.store('gave_mono_to(janice, phil)')
b = kb.bfs('john', 'phil'); b = list(b); assert len(b) == 2
assert len(b[0]) == 2; assert b[0][0] == ('gave_mono_to', 'jane')
assert b[0][1] == ('gave_mono_to', 'phil'); assert b[1][0] == ('gave_mono_to', 'jane');
assert b[1][1] == ('gave_mono_to', 'janice'); assert b[1][2] == ('gave_mono_to', 'phil')

print('All main KB tests passed.')


from collections import deque, defaultdict
from contextlib import contextmanager
import copy
import csv
import math
import os
import pickle
import random
import re
import sys

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
from scipy.special import expit
from sklearn.neighbors import NearestNeighbors
from sklearn.svm import SVC
from torch.utils.data import DataLoader
import torch
from tqdm import tqdm

from zincbase.graph.Edge import Edge
from zincbase.graph.Node import Node
from zincbase.logic.Goal import Goal
from zincbase.logic.Negative import Negative
from zincbase.logic.Term import Term
from zincbase.logic.Rule import Rule
from zincbase.logic.common import unify, process
from zincbase.nn.dataloader import NegDataset, TrainDataset, BidirectionalOneShotIterator
from zincbase.nn.rotate import KGEModel
from zincbase.utils.string_utils import strip_all_whitespace, split_to_parts, cleanse

class KB():
    """Knowledge Base Class

    >>> kb = KB()
    >>> kb.__class__
    <class 'zb.KB'>
    """
    def __init__(self):
        self.G = nx.MultiDiGraph()
        self.rules = []
        self._dont_propagate = False
        self._MAX_RECURSION = 1
        self._PROPAGATION_LIMIT = math.inf
        self._global_propagations = 0
        self._neg_examples = []
        self._entity2id = {}
        self._relation2id = {}
        self._encoded_triples = []
        self._encoded_neg_examples = []
        self._node_cache = {}
        self._edge_cache = {}
        self._kg_model = None
        self._knn = None
        self._knn_index = []
        self._cuda = False
        self.classifiers = {}

        self._model_name = None
        self._embedding_size = None
        self._gamma = None
        self._node_attributes = None
        self._pred_attributes = None
        self._attr_loss_to_graph_loss = None
        self._pred_loss_to_graph_loss = None

    def seed(self, seed):
        """Seed the RNGs for PyTorch, NumPy, and Python itself.

        :param int seed: random seed

        :Example:

        >>> KB().seed(555)
        """
        torch.random.manual_seed(seed)
        np.random.seed(seed)
        random.seed(seed)
    
    def set_recursion_limit(self, recursions):
        """Set a recursion limit for the KB. Complex arrangements
        can end up propagating back to themselves; permit this only
        a certain number of times. Also, `with kb.dont_propagate()`
        will prohibit any cycles, overriding this limit.

        This also sets the global Python recursion limit equivalently,
        up to a (high but often reasonable) maximum of 50,000. 
        (This may not be appropriate if you have very complex, memory-hungry
        watch functions; it may require some trial-and-error tweaking.)

        The default Zincbase recursion limit is 1 which is very low.
        Depending on your simulation, increasing this value may offer
        magnitudes higher performance.

        :param int recursions: Limit on recursions to same node within
        one update.

        :Example:

        >>> KB().set_recursion_limit(1e5)
        """
        limit = min(int(recursions), int(5e5))
        sys.setrecursionlimit(max(limit, 3000))
        self._MAX_RECURSION = recursions
    
    def set_propagation_limit(self, propagations):
        """Set a limit on the number of times one change to a node
        attribute may propagate to its neighbors, their neighbors, etc.
        If this limit is 0 it's equivalent to `with kb.dont_propagate():`
        By default, the limit is infinity (i.e. full global network effects)

        Consider using together with `kb.set_recursion_limit`.

        :param int propagations: Limit on propagations (ie, global network
        effects)

        :Example:
        >>> kb = KB()
        >>> kb.set_propagation_limit(1)
        """
        self._PROPAGATION_LIMIT = propagations

    def edge(self, sub, pred, ob):
        """Returns an edge and its attributes.

        :param str sub: Subject node/entity
        :param str pred: Predicate between subject and object
        :param str ob: Object node/entity

        :Example:

        >>> kb = KB()
        >>> kb.store('eats(tom, rice)')
        0
        >>> kb.edge('tom', 'eats', 'rice')
        tom___eats___rice
        >>> kb.edge('tom', 'eats', 'rice').alot = 1
        >>> kb.edge('tom', 'eats', 'rice').alot
        1
        """
        try:
            edge = self._edge_cache[(sub, pred, ob)]
        except:
            edge = Edge(self, sub, pred, ob)
            self._edge_cache[(sub, pred, ob)] = edge
        return edge
    
    @contextmanager
    def dont_propagate(self):
        self._dont_propagate = True
        yield self._dont_propagate
        self._dont_propagate = False

    def node(self, node_name):
        """Get a node, and its attributes, from the graph.

        :param str node_name: Name of the node
        :return: The node and its attributes.

        :Example:

        >>> kb = KB()
        >>> kb.store('eats(tom, rice)')
        0
        >>> kb.node('tom')
        tom
        >>> kb.node('tom').attrs
        {}
        >>> kb.node('tom').is_person = True
        >>> kb.node('tom').attrs
        {'is_person': True}"""
        node_name = str(node_name)
        try:
            node = self._node_cache[node_name]
        except:
            node = Node(self, node_name, self.G.nodes(data=True)[node_name])
            self._node_cache[node_name] = node
        return node

    def _valid_neighbors(self, node, reverse=False):
        if reverse:
            graph = self.G.reverse()
        else:
            graph = self.G
        neighbors = graph[node]
        return [x for x in neighbors.items()]
    
    def neighbors(self, node):
        """Return neighbors of node and predicates that connect them.

        :param str node: Name of the node
        :return: List[(node_name, List[predicate])]

        :Example:

        >>> kb = KB()
        >>> kb.store('knows(tom, shamala)')
        0
        >>> kb.neighbors('tom')
        [('shamala', [{'pred': 'knows'}])]"""
        neighbors = self._valid_neighbors(node)
        l = []
        for n in neighbors:
            how = []
            for rel in n[1]:
                how.append(n[1][rel])
            l.append((n[0], how))
        return l
    
    def filter(self, filter_condition, candidate_nodes=None):
        """Filter (ie query) nodes by attributes.

        :param function filter_condition: Test function
        :param List candidate_nodes: Nodes to test (optional; defaults to whole graph)

        :Example:

        >>> kb = KB()
        >>> kb.store('person(tom)')
        0
        >>> kb.node('tom').cats = 0
        >>> list(kb.filter(lambda x: x['cats'] < 1))
        [tom]"""
        if candidate_nodes is None:
            candidate_nodes = self.G.nodes
        for node in candidate_nodes:
            node = self.node(node)
            try:
                if filter_condition(node):
                    yield node
            except KeyError:
                # maybe node doesn't have the attr set
                pass


    def bfs(self, start_node, target_node, max_depth=10, reverse=False):
        """Find a path from start_node to target_node"""
        stack = [(start_node, 0, [])]
        answers = []
        while stack:
            node, depth, path = stack.pop(0)
            if depth >= max_depth:
                return answers
            for n, pred in self._valid_neighbors(node, reverse=reverse):
                if n == target_node:
                    for final_edge in pred:
                        yield path + [(pred[final_edge]['pred'], n)]
                else:
                    for edge in pred:
                        stack.append((n, depth+1, path + [(pred[edge]['pred'], n)]))
        return answers

    def add_node_to_trained_kg(self, sub, pred, ob):
        if (sub not in self._entity2id and ob not in self._entity2id) or (pred not in self._relation2id):
            raise Exception('Must have at least a known predicate and one of subject/object in the graph already.')
        known_sub = False
        if sub in self._entity2id:
            known_sub = True
        embeddings_copy = self._kg_model.entity_embedding.clone().detach().requires_grad_(True)
        new_embed = torch.zeros((1, embeddings_copy.shape[1]), requires_grad=False)
        if known_sub:
            nodes = self.query('{}({}, X)'.format(pred, sub))
            self._entity2id[ob] = len(self._entity2id)
        else:
            nodes = self.query('{}(X, {})'.format(pred, ob))
            self._entity2id[sub] = len(self._entity2id)
        for node in nodes:
            new_embed += self.get_embedding(node['X'])
        new_embed /= len(nodes)
        # TODO: Relations have embeddings also; add the relation embedding to new_embed each
        # time and average it.
        new_embed = new_embed.clone().detach().requires_grad_(True)
        self.store('{}({}, {})'.format(pred, sub, ob))
        self._kg_model.entity_embedding = torch.nn.Parameter(torch.cat((embeddings_copy, new_embed)))

    def create_multi_classifier(self, pred):
        """Build a classifier (SVM) for a predicate that can classify a subject, given a predicate, into
        one of the object entities from the KB that has that predicate relation. Automatically
        compensates for class imbalance.

        :Example:

        >>> kb = KB()
        >>> kb.from_csv('./assets/countries_s1_train.csv', delimiter='\\t')
        >>> kb.seed(555)
        >>> kb.build_kg_model(cuda=False, embedding_size=40)
        >>> kb.train_kg_model(steps=1000, batch_size=1, verbose=False)
        >>> _ = kb.create_multi_classifier('locatedin')
        >>> kb.multi_classify('philippines', 'locatedin')
        'south_eastern_asia'"""

        all_examples = list(self.query('{}(X, Y)'.format(pred)))
        Xs = []
        Ys = []
        indexes = list(set([x['Y'] for x in all_examples]))
        ratios = defaultdict(int)
        for example in all_examples:
            Xs.append(self.get_embedding(example['X']).cpu())
            Ys.append(indexes.index(example['Y']))
            ratios[indexes.index(example['Y'])] += 1
        Xs = np.reshape(np.stack(Xs), (-1, self.get_embedding(all_examples[0]['X']).shape[1]))
        Ys = np.stack(Ys)
        num_in_biggest_class = max(v for (k, v) in ratios.items())
        for ratio in ratios:
            ratios[ratio] = num_in_biggest_class / ratios[ratio]
        clf = SVC(gamma='auto', kernel='linear', class_weight=ratios)
        clf.fit(Xs, Ys)
        self.classifiers[pred] = (clf, indexes)
        return clf

    def multi_classify(self, subject, pred):
        """Predict `object` for subject according to the multi-classifer
        previously trained on `pred`."""

        clf, indexes = self.classifiers[pred]
        return indexes[int(clf.predict(np.reshape(self.get_embedding(subject).cpu(), (1, -1))))]

    def create_binary_classifier(self, pred, ob):
        """Creates a binary classifier (SVM) for `pred(?, ob)` using embeddings from the trained model.
        Automatically compensates for class imbalance.

        Follow it with `binary_classify(sub, pred, ob)` to predict whether the relation holds or not.

        May be useful because although the model can estimate a probability for (sub, pred, ob),
        what threshold should you use to decide what constitutes True vs False?

        :Example:

        >>> kb = KB()
        >>> kb.seed(555)
        >>> kb.from_csv('./assets/countries_s1_train.csv', delimiter='\\t')
        >>> kb.build_kg_model(cuda=False, embedding_size=100)
        >>> kb.train_kg_model(steps=2000, batch_size=1, verbose=False, neg_to_pos=4)
        >>> _ = kb.create_binary_classifier('locatedin', 'asia')
        >>> kb.binary_classify('india', 'locatedin', 'asia')
        True
        >>> kb.binary_classify('brazil', 'locatedin', 'asia')
        False"""
        all_examples = list(self.query('{}(X, Y)'.format(pred)))
        pos_examples = [self.get_embedding(x['X']) for x in all_examples if x['Y'] == ob]
        neg_examples = [self.get_embedding(x['X']) for x in all_examples if x['Y'] != ob]
        Xs = np.reshape(np.stack(pos_examples + neg_examples), (-1, pos_examples[0].shape[1]))
        Ys = np.stack([2 for x in pos_examples] + [1 for x in neg_examples])
        ratio = int(len(neg_examples) / len(pos_examples))
        clf = SVC(gamma='auto', kernel='linear', class_weight={2:min(ratio, 15)})
        clf.fit(Xs, Ys)
        self.classifiers[(pred, ob)] = clf
        return clf

    def binary_classify(self, subject, pred, ob):
        """Predict whether triple (sub, pred, ob) is true or not."""
        clf = self.classifiers[(pred, ob)]
        X = self.get_embedding(subject)
        pred = int(clf.predict(X))
        return pred == 2

    def save_all(self, dirname='.'):
        """Save current KB to the directory specified. Saves the (state dict of the) PyTorch \
        model as well, if it has been built.

        :param str dirname: Directory in which to save the files. Creates the directory \
        if it doesn't already exist."""
        if not os.path.exists(dirname):
            os.mkdir(dirname)
        if self._kg_model:
            torch.save(self._kg_model.state_dict(), os.path.join(dirname, 'pytorch_model.dict'))
        zb_dict = {
            'model_name': self._model_name,
            'entity2id': self._entity2id,
            'relation2id': self._relation2id,
            'encoded_triples': self._encoded_triples,
            'embedding_size': self._embedding_size,
            'gamma': self._gamma,
            'node_attributes': self._node_attributes,
            'pred_attributes': self._pred_attributes,
            'attr_loss_to_graph_loss': self._attr_loss_to_graph_loss,
            'pred_loss_to_graph_loss': self._pred_loss_to_graph_loss,
            'rules': self.rules
        }
        f = open(os.path.join(dirname, 'zb.pkl'), 'wb')
        pickle.dump(zb_dict, f)
        f.close()
        return True

    def load_all(self, dirname='.', cuda=False):
        """Load KB (and model, if it exists) from the specified directory.

        :param str dirname: Directory to load zb.pkl and (if present) pytorch_model.dict
        :param bool cuda: If the model exists, it will be loaded - specify if you want \
        it to be on the GPU."""

        with open(os.path.join(dirname, 'zb.pkl'), 'rb') as f:
            zb_dict = pickle.load(f)
        self._model_name = zb_dict['model_name']
        self._entity2id = zb_dict['entity2id']
        self._relation2id = zb_dict['relation2id']
        self._encoded_triples = zb_dict['encoded_triples']
        self._embedding_size = zb_dict['embedding_size']
        self._gamma = zb_dict['gamma']
        self._node_attributes = zb_dict['node_attributes']
        self._pred_attributes = zb_dict['pred_attributes']
        self._attr_loss_to_graph_loss = zb_dict['attr_loss_to_graph_loss']
        self._pred_loss_to_graph_loss = zb_dict['pred_loss_to_graph_loss']
        tmp_rules = zb_dict['rules']
        for rule in tmp_rules:
            self.store(str(rule.head))
        if os.path.exists(os.path.join(dirname, 'pytorch_model.dict')):
            self.build_kg_model(cuda, embedding_size=self._embedding_size, gamma=self._gamma,
            model_name=self._model_name, node_attributes=self._node_attributes,
            attr_loss_to_graph_loss=self._attr_loss_to_graph_loss,
            pred_loss_to_graph_loss=self._pred_loss_to_graph_loss,
            pred_attributes=self._pred_attributes)
            self._kg_model.load_state_dict(torch.load(os.path.join(dirname, 'pytorch_model.dict')))
        return True


    def build_kg_model(self, cuda=False, embedding_size=256, gamma=24, model_name='RotatE',
                    node_attributes=[], attr_loss_to_graph_loss=1.0, pred_loss_to_graph_loss=1.0,
                    pred_attributes=[]):
        """Build the dictionaries and KGE model

        :param list node_attributes: List of node attributes to include in the model. \
            If node doesn't possess the attribute, will be treated as zero. So far attributes \
        must be floats.
        :param list pred_attributes: List of predicate attributes to include in the model.
        :param float attr_loss_to_graph_loss: % to scale attribute loss against graph loss. \
        0 would only take into account graph loss, math.inf would only take into account attr loss."""
        # TODO refactor this so there's a separate dict of node + pred attrs; they don't have
        # to be part of the triple.

        self._gamma = gamma
        self._embedding_size = embedding_size
        self._model_name = model_name
        self._attr_loss_to_graph_loss = attr_loss_to_graph_loss
        self._pred_loss_to_graph_loss = pred_loss_to_graph_loss
        self._node_attributes = node_attributes
        self._pred_attributes = pred_attributes

        triples = self.to_triples(data=True)
        for i, triple in enumerate(triples):
            if triple[0] not in self._entity2id:
                self._entity2id[triple[0]] = len(self._entity2id)
        for i, triple in enumerate(triples):
            if triple[1] not in self._relation2id:
                self._relation2id[triple[1]] = len(self._relation2id)
        curlen = len(self._entity2id)
        j = 0
        for i, triple in enumerate(triples):
            if triple[2] not in self._entity2id:
                self._entity2id[triple[2]] = curlen + j
                j += 1
        self._encoded_triples = []
        for triple in triples:
            # TODO: attribute must be a float; for a dictionary encoding of them (for categoricals)
            attrs = []
            for attribute in node_attributes:
                attr = float(triple[3].get(attribute, 0.0))
                attrs.append(attr)
            for pred_attr in pred_attributes:
                if pred_attr == 'truthiness':
                    default_value = 1.
                else:
                    default_value = 0.
                attr = float(triple[4].get(pred_attr, default_value))
                attrs.append(attr)
            if len(triple) == 7 and triple[6]:
                true = 1. # it's a false fact; negative example TODO rename from 'true'!
            else:
                true = 0.
            self._encoded_triples.append((self._entity2id[triple[0]], self._relation2id[triple[1]], self._entity2id[triple[2]],
                                        attrs, true))
        for neg_example in self._neg_examples:
            self._encoded_neg_examples.append((self._entity2id[neg_example.head], self._relation2id[neg_example.pred], self._entity2id[neg_example.tail]))
        dee = False; dre = False
        if model_name == 'ComplEx':
            dee = True
            dre = True
        if model_name == 'RotatE':
            dee = True
            dre = False
        if cuda:
            device = 'cuda'
        else:
            device = 'cpu'
        self._kg_model = KGEModel(model_name=model_name,
                             nentity=len(self._entity2id),
                             nrelation=len(self._relation2id),
                             hidden_dim=embedding_size,
                             gamma=gamma,
                             double_entity_embedding=dee,
                             double_relation_embedding=dre,
                             node_attributes=node_attributes,
                             pred_attributes=pred_attributes,
                             attr_loss_to_graph_loss=attr_loss_to_graph_loss,
                             pred_loss_to_graph_loss=pred_loss_to_graph_loss,
                             device=device)
        if cuda:
            self._cuda = True
            self._kg_model = self._kg_model.cuda()

    def train_kg_model(self, steps=1000, batch_size=512, lr=0.001,
                       reencode_triples=False, neg_to_pos=128,
                       neg_ratio=1., verbose=True):
        """Train a KG model on the KB.

        :param int steps: Number of training steps
        :param int batch_size: Batch size for training
        :param float lr: Initial learning rate for Adam optimizer
        :param bool reencode_triples: If a node has been added since last training, set this to True
        :param int neg_to_pos: Ratio of generated negative samples to real positive samples
        :param float neg_ratio: How often real/inputted negative examples should appear, vs real pos + generated neg. Smaller (>0) means more often.
        """
        if reencode_triples:
            # TODO: this is not encoding attributes as well, yet.
            triples = self.to_triples(data=True)
            self._encoded_triples = []
            for triple in triples:
                self._encoded_triples.append((self._entity2id[triple[0]], self._relation2id[triple[1]], self._entity2id[triple[2]]))

        nentity = len(self._entity2id)
        nrelation = len(self._relation2id)
        train_dataloader_head = DataLoader(
            TrainDataset(self._encoded_triples, nrelation, neg_to_pos, 'head-batch'),
            batch_size=batch_size,
            shuffle=True,
            num_workers=1,
            collate_fn=TrainDataset.collate_fn)
        train_dataloader_tail = DataLoader(
            TrainDataset(self._encoded_triples, nrelation, neg_to_pos, 'tail-batch'),
            batch_size=batch_size,
            shuffle=True,
            num_workers=1,
            collate_fn=TrainDataset.collate_fn)
        if len(self._neg_examples):
            neg_dataloader = DataLoader(
                NegDataset(self._encoded_neg_examples),
                batch_size=batch_size,
                shuffle=True,
                num_workers=1,
                collate_fn=TrainDataset.collate_fn)
            neg_ratio = int(neg_ratio * (len(self._encoded_triples) / len(self._neg_examples)))
            neg_ratio = max(neg_ratio, 1e-4)
            train_iterator = BidirectionalOneShotIterator(train_dataloader_head, train_dataloader_tail, neg_dataloader, neg_ratio)
        else:
            train_iterator = BidirectionalOneShotIterator(train_dataloader_head, train_dataloader_tail)
        optimizer = torch.optim.Adam(filter(lambda p: p.requires_grad, self._kg_model.parameters()), lr=lr)

        self._kg_model.train()
        if verbose:
            it = tqdm(range(0, steps))
        else:
            it = range(0, steps)
        for step in it:
            log = self._kg_model.train_step(self._kg_model, optimizer, train_iterator, {'cuda': self._cuda})
            if verbose and step % 100 == 0:
                print(log)
        self._kg_model.eval()

    def estimate_triple_prob(self, sub, pred, ob):
        """Estimate the probability of the triple (sub, pred, ob) according to the trained model."""

        # TODO: Should be prolog style
        if not self._kg_model:
            raise Exception('Must build and train the model first')
        tensor = torch.tensor([[self._entity2id[sub], self._relation2id[pred], self._entity2id[ob]]])
        if self._cuda:
            tensor = tensor.cuda()
        logit, _ = self._kg_model(tensor, attributes=False, predict_only=True)
        return round(expit(float(logit)), 4)

    def estimate_triple_prob_with_attrs(self, sub, pred, ob, pred_prop):
        # TODO: Should be prolog style
        if not self._kg_model:
            raise Exception('Must build and train the model first')
        tensor = torch.tensor([[self._entity2id[sub], self._relation2id[pred], self._entity2id[ob]]])
        if self._cuda:
            tensor = tensor.cuda()
        logit, _ = self._kg_model(tensor, attributes=True, predict_pred_prop=pred_prop, predict_only=True)
        return round(expit(float(logit)), 4)

    def get_embedding(self, entity):
        index = torch.LongTensor([self._entity2id[entity]])
        if self._cuda:
            index = index.cuda()
        return torch.index_select(self._kg_model.entity_embedding, dim=0, index=index).detach()

    def fit_knn(self, entities=None):
        """Fit an unsupervised sklearn kNN to the embeddings of entities.

        :param list entities: The entities that should be part of the kNN. Defaults to all if not specified"""
        self._knn_index = []
        if not entities:
            entities = [e for e in self._entity2id]
        encoded_entities = []
        for e in entities:
            encoded_entities.append(self._entity2id[e])
            self._knn_index.append(e)
        index = torch.LongTensor(encoded_entities)
        if self._cuda:
            index = index.cuda()
        embeddings = torch.index_select(self._kg_model.entity_embedding, dim=0, index=index).detach().cpu()
        self._knn = NearestNeighbors(n_neighbors=4, algorithm='kd_tree').fit(embeddings)

    def get_nearest_neighbors(self, entity, k=1):
        """Get the nearest neighbors to entity (embedding), according to the previously fit knn.

        :param str entity: An entity
        :param int k: How many neighbors
        """
        embedding = self.get_embedding(entity)
        embedding = embedding.cpu() # no cuda for sklearn
        distances, indices = self._knn.kneighbors(embedding, n_neighbors=k)
        borgs = []
        distances = distances[0]
        indices = indices[0]
        for i in range(len(distances)):
            borgs.append({'distance': round(distances[i], 4), 'entity': self._knn_index[int(indices[i])]})
        return borgs

    @property
    def entities(self):
        """All the entities in the KB.

        :returns generator: Generator of all the entities"""
        return self._entity2id.keys()

    @property
    def predicates(self):
        """All the predicates (aka relations) in the KB.

        :returns generator: Generator of all the predicates"""
        return self._relation2id.keys()

    def get_most_likely(self, sub, pred, ob, candidates=None, k=1):
        """Return the k most likely triples to satisfy the input triple. One of \
        sub, pred, or ob may be '?'.

        :param list<str> candidates: Candidate entities/predicates. If None or not specified, this function \
        will generate possible candidates from the rest of the triple.
        :param int k: The k in top k.

        :Example:

        >>> kb = KB()
        >>> kb.from_csv('./assets/countries_s1_train.csv', delimiter='\\t')
        >>> kb.seed(555)
        >>> kb.build_kg_model(cuda=False, embedding_size=100)
        >>> kb.train_kg_model(steps=2000, batch_size=2, verbose=False, neg_to_pos=4)
        >>> kb.get_most_likely('austria', 'neighbor', '?', k=2) # doctest:+ELLIPSIS
        [{'prob': 0.9673, 'triple': ('austria', 'neighbor', 'germany')}, {'prob': 0.9656, 'triple': ('austria', 'neighbor', 'liechtenstein')}]
        >>> kb.get_most_likely('?', 'neighbor', 'austria', candidates=list(kb.entities), k=2)
        [{'prob': 0.9467, 'triple': ('slovenia', 'neighbor', 'austria')}, {'prob': 0.94, 'triple': ('liechtenstein', 'neighbor', 'austria')}]
        >>> kb.get_most_likely('austria', '?', 'germany', k=3)
        [{'prob': 0.9673, 'triple': ('austria', 'neighbor', 'germany')}, {'prob': 0.664, 'triple': ('austria', 'locatedin', 'germany')}]"""
        
        reverse_lookup = {}
        possibles = []
        orig_sub = sub
        orig_ob = ob
        if not candidates:
            if pred == '?':
                candidates = self.predicates
            else:
                if sub == '?':
                    sub = 'X'
                    ob = 'Y'
                else:
                    ob = 'X'
                    sub = 'Y'
                candidates = self.query('{}({}, {})'.format(pred, sub, ob))
                candidates = list(set([x['X'] for x in candidates]))
        for cand in candidates:
            if pred == '?':
                reverse_lookup[self._relation2id[cand]] = cand
                possibles.append([self._entity2id[sub], self._relation2id[cand], self._entity2id[ob]])
            else:
                reverse_lookup[self._entity2id[cand]] = cand
                if orig_sub == '?':
                    possibles.append([self._entity2id[cand], self._relation2id[pred], self._entity2id[orig_ob]])
                else:
                    possibles.append([self._entity2id[orig_sub], self._relation2id[pred], self._entity2id[cand]])
        possibles_tensor = torch.tensor(possibles)
        if self._cuda:
            possibles_tensor = possibles_tensor.cuda()
        out, _ = self._kg_model(possibles_tensor, predict_only=True)
        k = min(out.size(0), k)
        answers = torch.topk(out, k=k, dim=0)
        probs = answers[0]
        indexes = answers[1]
        retvals = []
        for i in range(len(indexes)):
            if pred == '?':
                orig = reverse_lookup[possibles[int(indexes[i])][1]]
                triple = (sub, orig, ob)
            elif orig_sub == '?':
                orig = reverse_lookup[possibles[int(indexes[i])][0]]
                triple = (orig, pred, orig_ob)
            else:
                orig = reverse_lookup[possibles[int(indexes[i])][2]]
                triple = (orig_sub, pred, orig)
            retvals.append({'prob': round(expit(float(probs[i])), 4), 'triple': triple})
        return retvals

    def _search(self, term):
        head_goal = Goal(Rule("x(y):-x(y)"))
        head_goal.rule.goals = [term]
        queue = deque([head_goal])
        iterations = 0
        max_iterations = max(100, (len(self.rules) + 1) ** 1.5)
        while queue and iterations < max_iterations:
            iterations += 1
            c = queue.popleft()
            if c.idx >= len(c.rule.goals):
                if not c.parent:
                    if c.bindings:
                        new_binding = {k:str(v) for (k, v) in c.bindings.items()}
                        yield new_binding
                    else:
                        yield True
                    continue
                parent = copy.deepcopy(c.parent)
                unify(c.rule.head, c.bindings, parent.rule.goals[parent.idx], parent.bindings)
                parent.idx += 1
                queue.append(parent)
                continue
            term = c.rule.goals[c.idx]
            pred = term.pred
            for rule in self.rules:
                if rule.head.pred != term.pred:
                    continue
                if len(rule.head.args) != len(term.args):
                    continue
                child = Goal(rule, c)
                ans = unify(term, c.bindings, rule.head, child.bindings)
                if ans:
                    queue.append(child)

    def delete_rule(self, rule_idx):
        """Delete a rule from the KB.

        :param rule_idx: The index of the rule in the KB. Returned when the rule was added. May be int (if it \
        was a real rule) or str (if it was a negative example - preceded by ~).

        :Example:

        >>> kb = KB()
        >>> kb.store('a(a)')
        0
        >>> kb.delete_rule(0)
        True
        """
        try:
            if isinstance(rule_idx, str) and rule_idx[0] == '~':
                rule_idx = int(rule_idx[1:])
                self._neg_examples.pop(rule_idx)
                return True
            self.rules.pop(rule_idx)
            return True
        except:
            return False

    def plot(self, density=1.0):
        """Plots a network diagram from (triple) nodes and edges in the KB.

        :param float density: Probability (0-1) that a given edge will be plotted, \
        useful to thin out dense graphs for visualization."""
        edgelist = [e for e in self.G.edges(data=True) if random.random() < density]
        newg = nx.DiGraph(edgelist)
        pos = nx.spring_layout(newg)
        plt.figure(1,figsize=(12,12))
        nx.draw_networkx_nodes(newg, pos, node_size=200)
        nx.draw_networkx_edges(newg, pos, edgelist=edgelist, width=1, font_size=8)
        nx.draw_networkx_labels(newg, pos, font_size=10, font_family='sans-serif')
        nx.draw_networkx_edge_labels(newg, pos)
        plt.axis('off')
        plt.show()

    def solidify(self, predicate):
        """Query the KB (with Prolog) and 'solidify' facts in the KB, making them part
        of the graph, so that the NN can be trained.

        :param str predicate: A predicate (that's a rule not a fact otherwise what's the point)

        :Example:

        >>> kb = KB()
        >>> kb.store('is(tom, human)')
        0
        >>> kb.store('has_part(shamala, head)')
        1
        >>> kb.store('is(X, human) :- has_part(X, head)')
        2
        >>> next(kb.query('is(tom, human)'))
        True
        >>> kb.to_triples()
        [('tom', 'is', 'human'), ('shamala', 'has_part', 'head')]
        >>> kb.solidify('is')
        1
        >>> kb.to_triples()
        [('tom', 'is', 'human'), ('shamala', 'has_part', 'head'), ('shamala', 'is', 'human')]
        """
        answers = self.query('{}(X, Y)'.format(predicate))
        i = 0
        rule_strings = [str(x) for x in self.rules]
        for a in answers:
            as_string = '{}({}, {})'.format(predicate, a['X'], a['Y'])
            if not as_string in rule_strings:
                i += 1
                self.store(as_string)
        return i

    def query(self, statement):
        """Query the KB.

        :param str statement: A rule to query on.
        :return: Generator of alternative bindings to variables that match the query

        :Example:

        >>> kb = KB()
        >>> kb.store('a(a)')
        0
        >>> kb.query('a(X)') #doctest: +ELLIPSIS
        <generator object KB._search at 0x...>
        >>> list(kb.query('a(X)'))
        [{'X': 'a'}]"""
        return self._search(Term(strip_all_whitespace(statement)))

    def store(self, statement, node_attributes=[], edge_attributes={}):
        """Store a fact/rule in the KB

        It is possible to store 'false' facts (negative examples) by preceding the predicate with a tilde (~).
        In this case, they do not come out in the graph and cannot be queried, but may
        assist when building the model.

        :param str statement: Fact or rule to store in the KB.
        :param list<dict> node_attributes: List with the same number of elements as there \
        are nodes in the statement, with each element being a \
        dict of items to set on the nodes.
        :param dict edge_attributes: Dictionary of attributes to set on the edge. May \
        include truthiness which, if < 0, automatically makes the rule a negative example.
        :return: the id of the fact/rule

        :Example:

        >>> kb = KB()
        >>> kb.store('a(a)')
        0
        >>> kb.store('node(x)', node_attributes=[{'node_number': 1}])
        1
        >>> list(kb.query('node(What)'))
        [{'What': 'x'}]"""
        statement = strip_all_whitespace(statement)
        if 'truthiness' in edge_attributes and edge_attributes['truthiness'] < 0:
            if statement[0] != '~':
                statement = '~' + statement
        if statement[0] == '~':
            triple = split_to_parts(statement[1:])
            if not triple[0] in self._entity2id:
                self._entity2id[triple[0]] = len(self._entity2id)
            if not triple[1] in self._relation2id:
                self._relation2id[triple[1]] = len(self._relation2id)
            if not triple[2] in self._entity2id:
                self._entity2id[triple[2]] = len(self._entity2id)
            self._neg_examples.append(Negative(statement[1:]))
            return '~' + str(len(self._neg_examples) - 1)
        self.rules.append(Rule(statement, kb=self))
        if edge_attributes:
            parts = split_to_parts(statement)
            if parts[2] is not None:
                for idx, edge in self.G[parts[0]][parts[2]].items():
                    if edge['pred'] == parts[1]:
                        nx.set_edge_attributes(self.G, {(parts[0], parts[2], idx): edge_attributes})
        if node_attributes:
            parts = split_to_parts(statement)
            nx.set_node_attributes(self.G, {parts[0]: node_attributes[0]})
            if parts[2] is not None:
                nx.set_node_attributes(self.G, {parts[2]: node_attributes[1]})
        return len(self.rules) - 1

    def to_tensorboard_projector(self, embeddings_filename, labels_filename, filter_fn=None):
        """Convert the KB's trained embeddings to 2 files suitable for \
        https://projector.tensorflow.org. This outputs only entity embeddings, \
        not relation embeddings, a visualization of which may not be interpretable.

        :param str embeddings_filename: Filename to output embeddings to, tsv format.
        :param str labels_filename: Filename to output labels to, one label per row.
        :param function filter_fn: Only include the embeddings/labels for which filter_fn(label) returns True"""

        embeddings = self._kg_model.entity_embedding.detach().cpu().numpy()
        if not filter_fn:
            filter_fn = lambda x: True
        z = [e for e in zip(embeddings, self.entities) if filter_fn(e[1])]
        embeddings, labels = list(zip(*z))
        np.savetxt(embeddings_filename, embeddings, delimiter='\t')
        labels_file = open(labels_filename, 'w')
        for ent in labels:
            labels_file.write(ent + '\n')
        labels_file.close()
        return True

    def to_triples(self, data=False):
        """Convert all facts in the KB to a list of triples, each of length 3
        (or 4 if data=True).
        Any fact that is not arity 2 will be ignored.

        :Note: While the Prolog style representation uses `pred(subject, object)`, \
        the triple representation is `(subject, pred, object)`.

        :param bool data: Whether to return subject, predicate and object \
        attributes as elements 4, 5, and 6 of the triple. The 7th element of the \
        triple is usually False, but is True when the fact/triple is a negative example.
        :return: list of triples (tuples of length 3 or 7 if data=True)

        :Example:

        >>> kb = KB()
        >>> kb.store('a(b, c)')
        0
        >>> kb.to_triples()
        [('b', 'a', 'c')]
        >>> kb.store('a(a)')
        1
        >>> kb.to_triples()
        [('b', 'a', 'c')]
        >>> kb.node('b').an_attribute = 'xyz'
        >>> kb.to_triples()
        [('b', 'a', 'c')]
        >>> kb.to_triples(data=True)
        [('b', 'a', 'c', {'an_attribute': 'xyz'}, {}, {}, False)]"""
        triples = []
        neg_examples = [str(x) for x in self._neg_examples]
        for r in self.rules:
            if not r.goals:
                if len(r.head.args) == 2:
                    subject = str(r.head.args[0])
                    subject = subject[0].lower() + subject[1:]
                    object_ = str(r.head.args[1])
                    object_ = object_[0].lower() + object_[1:]
                    if data:
                        edge = self.edge(subject, r.head.pred, object_)
                        truthiness = edge.get('truthiness', False)
                        if (truthiness and truthiness < 0) or str(r) in neg_examples:
                            is_neg = True
                        else:
                            is_neg = False
                        triples.append((subject, r.head.pred, object_,
                            self.node(subject).attrs,
                            edge.attrs,
                            self.node(object_).attrs,
                            is_neg
                        ))
                    else:
                        triples.append((subject, r.head.pred, object_))
        return triples

    def from_triples(self, triples):
        """Stores facts from a list of tuples into the KB.

        :param list triples: List of tuples each of the form `(subject, pred, object)`

        :Example:

        >>> kb = KB()
        >>> kb.from_triples([('b', 'a', 'c')])
        >>> len(list(kb.query('a(b, c)')))
        1"""
        for (u, p, v) in triples:
            self.store('{}({},{})'.format(p, u, v))

    def from_csv(self, csvfile, header=None, start=0, size=None, delimiter=','):
        with open(csvfile) as f:
            reader = csv.reader(f, delimiter=delimiter)
            i = 0
            if header:
                next(reader, None)
                i = 1
            while i < start:
                i += 1
                next(reader, None)
            i = 0
            for row in reader:
                pred = cleanse(row[1])
                sub = cleanse(row[0])
                ob = cleanse(row[2])
                if not (sub.replace('_','').isalnum() and ob.replace('_','').isalnum()):
                    continue
                node_attributes = [
                    {'_': {'real_name': row[0]}},
                    {'_': {'real_name': row[2]}}
                ]
                edge_attributes = {
                    '_': {'real_name': row[1]}
                }
                self.store('{}({},{})'.format(pred, sub, ob), node_attributes=node_attributes, edge_attributes=edge_attributes)
                i += 1
                if size and i > size:
                    break



"""Under-the-hood details of ZincBase's Prolog-like implementation"""

import copy

from zincbase.logic.Term import Term
from zincbase.utils.type_checks import isVar, isAtom

def unify(src, src_bindings, dest, dest_bindings):
    if src.pred == '_' or dest.pred == '_':
        return True
    if isVar(src):
        tmp_src = process(src, src_bindings)
        if not tmp_src:
            return True
        else:
            return unify(tmp_src, src_bindings, dest, dest_bindings)
    if isVar(dest):
        tmp_dest = process(dest, dest_bindings)
        if tmp_dest:
            return unify(src, src_bindings, tmp_dest, dest_bindings)
        else:
            dest_bindings[dest.pred] = process(src, src_bindings)
            return True
    if len(src.args) != len(dest.args):
        return False
    elif src.pred != dest.pred:
        return False
    else:
        dest_bindings_copy = copy.deepcopy(dest_bindings)
        for i in range(len(src.args)):
            if not unify(src.args[i], src_bindings, dest.args[i], dest_bindings_copy):
                return False
        dest_bindings.update(dest_bindings_copy)
        return True

def process(term, bindings, graph=None):
    if isAtom(term):
        return term
    if isVar(term):
        ans = bindings.get(term.pred, None)
        if not ans:
            return None
        else:
            return process(ans, bindings)
    args = []
    for arg in term.args:
        a = process(arg, bindings)
        if not a:
            return None
        args.append(a)
    return Term(term.pred, args)



"""A base unit for ZincBase's Prolog-like implementation of 'facts'"""

from zincbase.utils.string_utils import split_on

class Term:
    def __init__(self, expr, args=None, kb=None):
        if args:
            self.pred = expr
            self.args = args
        elif expr[-1] == ']':
            arr = split_on(expr[1:-1], ',')
            headtail = split_on(expr[1:-1], '|')
            if len(headtail) > 1:
                self.args = [Term(f, kb=kb) for f in headtail]
                self.pred = '__list__'
            else:
                arr.reverse()
                first = Term('__list__', [])
                for part in arr:
                    first = Term('__list__', [Term(part, kb=kb), first], kb=kb)
                self.pred = first.pred
                self.args = first.args
        elif expr[-1] == ')':
            sub_exprs = split_on(expr, '(', all=False)
            if len(sub_exprs) != 2:
                raise Exception('Syntax error')
            self.args = [Term(sub_expr, kb=kb) for sub_expr in split_on(sub_exprs[1][:-1], ',')]
            self.pred = sub_exprs[0]
        else:
            self.pred = expr
            self.args = []

        if kb is not None:
            for i, arg in enumerate(self.args):
                if arg:
                    added_node_1 = False
                    if not kb.G.has_node(str(arg)):
                        kb.G.add_node(str(arg))
                        added_node_1 = True
                    for arg2 in self.args[i+1:]:
                        added_node_2 = False
                        if not kb.G.has_node(str(arg2)):
                            kb.G.add_node(str(arg2))
                            added_node_2 = True
                        kb.G.add_edge(str(arg), str(arg2), pred=self.pred)
                        if added_node_1:
                            node = kb.node(str(arg2))
                            try:
                                if not kb._dont_propagate:
                                    node._new_neighbor_fn(str(arg))
                            except Exception as e:
                                pass
                        if added_node_2:
                            node = kb.node(str(arg))
                            try:
                                if not kb._dont_propagate:
                                    node._new_neighbor_fn(str(arg2))
                            except Exception as e:
                                pass


    def __repr__(self):
        if self.pred == '__list__':
            if not self.args:
                return '[]'
            first = self.args[1]
            if first.pred == '__list__' and first.args == []:
                return '[{}]'.format(str(self.args[0]))
            elif first.pred == '__list__':
                return '[{},{}]'.format(str(self.args[0]), str(self.args[1])[1:-1])
            else:
                return '[{}|{}]'.format(str(self.args[0]), str(self.args[1]))
        elif self.args:
            return '{}({})'.format(self.pred, ', '.join(map(str,self.args)))
        else:
            return self.pred



import re

def strip_all_whitespace(line):
    return line.replace(' ', '').strip()

def split_on(line, separator, all=True):
    nesting = 0
    separator_length = len(separator)
    if not line:
        return []
    for i in range(len(line)):
        c = line[i]
        if nesting <= 0 and line[i:i + separator_length] == separator:
            if all:
                return [line[:i]] + split_on(line[i + separator_length:], separator)
            else:
                return [line[:i], line[i + separator_length:]]
        if c in ('[', '('):
            nesting += 1
        elif c in (']', ')'):
            nesting -= 1
            continue
    return [line]

def split_to_parts(line):
    sub_exprs = split_on(line, '(', all=False)
    if len(sub_exprs) != 2:
        raise Exception('Syntax error')
    atoms = split_on(sub_exprs[1][:-1], ',')
    return (atoms[0], sub_exprs[0], atoms[1] if len(atoms) > 1 else None)

def cleanse(line):
    line = re.sub('[ ./()-]', '_', line)
    return line[0].lower() + line[1:]


import eventlet
from flask import Flask
from flask_socketio import SocketIO, send

app = Flask(__name__, static_url_path='/')

@app.route('/')
def index():
    return app.send_static_file('index.html')

def serve(args):
    eventlet.monkey_patch()
    print('Starting websocket server...')
    
    socketio = SocketIO(app, message_queue=f"redis://{args.redis}", cors_allowed_origins='*')
    
    @socketio.on('message')
    def handle_message(message):
        pass

    @socketio.on('connect')
    def handle_connect():
        pass

    socketio.run(app)
    return app


import math

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from torch.utils.data import DataLoader

class KGEModel(nn.Module):
    def __init__(self, model_name, nentity, nrelation, hidden_dim, gamma,
                 double_entity_embedding=False, double_relation_embedding=False,
                 node_attributes=[], pred_attributes=[],
                 attr_loss_to_graph_loss=1.0, pred_loss_to_graph_loss=1.0,
                 device='cuda'):
        super(KGEModel, self).__init__()
        self.model_name = model_name
        self.nentity = nentity
        self.nrelation = nrelation
        self.hidden_dim = hidden_dim
        self.epsilon = 2.0
        self.attr_loss_to_graph_loss = attr_loss_to_graph_loss
        self.pred_loss_to_graph_loss = pred_loss_to_graph_loss
        self.device = device

        self.gamma = nn.Parameter(torch.Tensor([gamma]), requires_grad=False)

        self.embedding_range = nn.Parameter(
            torch.Tensor([(self.gamma.item() + self.epsilon) / hidden_dim]),
            requires_grad=False)

        self.entity_dim = hidden_dim*2 if double_entity_embedding else hidden_dim
        self.relation_dim = hidden_dim*2 if double_relation_embedding else hidden_dim

        self.entity_embedding = nn.Parameter(torch.zeros(nentity, self.entity_dim))
        nn.init.uniform_(tensor=self.entity_embedding,
            a=-self.embedding_range.item(),
            b=self.embedding_range.item())

        self.relation_embedding = nn.Parameter(torch.zeros(nrelation, self.relation_dim))
        nn.init.uniform_(tensor=self.relation_embedding,
            a = -self.embedding_range.item(),
            b = self.embedding_range.item())

        self.node_attributes = node_attributes
        self.num_node_attributes = len(node_attributes)

        self.final_layer_size = self.entity_dim * self.num_node_attributes
        if self.num_node_attributes:
            self.attribute_layer = nn.Linear(self.final_layer_size, self.num_node_attributes)
            self.attribute_layer.weight.requires_grad = False
            self.attribute_layer.bias.requires_grad = False
            self.attribute_layer.to(self.device)

        self.pred_attributes = pred_attributes
        self.num_pred_attributes = len(pred_attributes)

        if self.num_pred_attributes:
            self.pred_layer = nn.Linear((2 * self.entity_dim) + self.relation_dim, self.num_pred_attributes)
            self.pred_layer.weight.requires_grad = False
            self.pred_layer.bias.requires_grad = False
            self.pred_layer.to(self.device)

        self.attr_loss_fn = nn.SmoothL1Loss()
        self.nonlinearity = torch.tanh # Cannot use relu since layers non-trainable: could start and stay negative only

        if model_name not in ['ComplEx', 'RotatE']:
            raise ValueError('model {} not supported'.format(model_name))

    def run_embedding(self, embedding, attribute_name):
        x = self.attribute_layer(embedding.repeat(repeats=(1, self.num_node_attributes, 1)).flatten())
        x = self.nonlinearity(x)
        return x[self.node_attributes.index(attribute_name)].item()

    def forward(self, sample, mode='single', attributes=True, predict_pred_prop=False, predict_only=False):
        """A single forward pass"""
        if mode == 'single':
            batch_size, negative_sample_size = sample.size(0), 1

            head = torch.index_select(
                self.entity_embedding,
                dim=0,
                index=sample[:,0]
            ).unsqueeze(1)

            relation = torch.index_select(
                self.relation_embedding,
                dim=0,
                index=sample[:,1]
            ).unsqueeze(1)

            tail = torch.index_select(
                self.entity_embedding,
                dim=0,
                index=sample[:,2]
            ).unsqueeze(1)

            attr_node = sample[:, 3:3 + self.num_node_attributes]
            attr_node = attr_node.to(torch.float)
            attr_pred = sample[:, 3+self.num_node_attributes:]
            attr_pred = attr_pred.to(torch.float)

            true = sample[:, -1]

        elif mode == 'head-batch':

            tail_part, head_part = sample
            attr_node = head_part[:, 3:3 + self.num_node_attributes]
            attr_pred = head_part[:, 3 + self.num_node_attributes:]
            true = tail_part[:, -1]
            head_part = head_part[:, :3]
            tail_part = tail_part[:, :3]
            batch_size, negative_sample_size = head_part.size(0), head_part.size(1)

            attr_node = attr_node.to(torch.float)
            attr_pred = attr_pred.to(torch.float)
            head_part = head_part.to(torch.long)
            tail_part = tail_part.to(torch.long)

            head = torch.index_select(
                self.entity_embedding,
                dim=0,
                index=head_part.view(-1)
            ).view(batch_size, negative_sample_size, -1)

            relation = torch.index_select(
                self.relation_embedding,
                dim=0,
                index=tail_part[:, 1]
            ).unsqueeze(1)

            tail = torch.index_select(
                self.entity_embedding,
                dim=0,
                index=tail_part[:, 2]
            ).unsqueeze(1)

        elif mode == 'tail-batch':
            head_part, tail_part = sample
            batch_size, negative_sample_size = tail_part.size(0), tail_part.size(1)

            attr_node = head_part[:, 3:3 + self.num_node_attributes]
            attr_pred = head_part[:, 3 + self.num_node_attributes:]
            head_part = head_part.to(torch.long)
            tail_part = tail_part.to(torch.long)
            attr_node = attr_node.to(torch.float)
            attr_pred = attr_pred.to(torch.float)

            head = torch.index_select(
                self.entity_embedding,
                dim=0,
                index=head_part[:, 0]
            ).unsqueeze(1)

            relation = torch.index_select(
                self.relation_embedding,
                dim=0,
                index=head_part[:, 1]
            ).unsqueeze(1)

            tail = torch.index_select(
                self.entity_embedding,
                dim=0,
                index=tail_part.view(-1)
            ).view(batch_size, negative_sample_size, -1)

            true = head_part[:, -1]

        elif mode == 'neg':
            head = torch.index_select(
                self.entity_embedding,
                dim=0,
                index=sample[:,0]
            ).unsqueeze(1)

            relation = torch.index_select(
                self.relation_embedding,
                dim=0,
                index=sample[:,1]
            ).unsqueeze(1)

            tail = torch.index_select(
                self.entity_embedding,
                dim=0,
                index=sample[:,2]
            ).unsqueeze(1)

            true = sample[:, -1]

        model_func = {
            'ComplEx': self.ComplEx,
            'RotatE': self.RotatE
        }

        score = model_func[self.model_name](head, relation, tail, mode)
        if not predict_only and mode != 'neg':
            true = (1 - true).unsqueeze(dim=-1).to(torch.float32).to(self.device)
            score = score * true

        if mode == 'neg':
            score = -score

        if predict_pred_prop:
            whole = torch.cat((head.squeeze(), relation.squeeze(), tail.squeeze()), dim=-1)
            x = self.pred_layer(whole)
            x = self.nonlinearity(x)
            return x[self.pred_attributes.index(predict_pred_prop)].item(), None

        if not attributes:
            return score, None

        attr_loss = torch.tensor(0, dtype=torch.float, device=self.device)

        if mode == 'single':
            if self.num_node_attributes:
                big_head = head.repeat(repeats=(1, self.num_node_attributes, 1))
                attr_hat = self.attribute_layer(big_head.flatten().view(-1, self.final_layer_size))
                attr_hat = self.nonlinearity(attr_hat)
                attr_loss += self.attr_loss_to_graph_loss * self.attr_loss_fn(attr_hat, attr_node)
            if self.num_pred_attributes:
                whole = torch.cat((head.squeeze(), relation.squeeze(), tail.squeeze()), dim=-1)
                attr_hat = self.pred_layer(whole)
                attr_hat = self.nonlinearity(attr_hat)
                attr_loss += self.pred_loss_to_graph_loss * self.attr_loss_fn(attr_hat, attr_pred)

        return score, attr_loss

    def ComplEx(self, head, relation, tail, mode):
        re_head, im_head = torch.chunk(head, 2, dim=2)
        re_relation, im_relation = torch.chunk(relation, 2, dim=2)
        re_tail, im_tail = torch.chunk(tail, 2, dim=2)

        if mode == 'head-batch':
            re_score = re_relation * re_tail + im_relation * im_tail
            im_score = re_relation * im_tail - im_relation * re_tail
            score = re_head * re_score + im_head * im_score
        else:
            re_score = re_head * re_relation - im_head * im_relation
            im_score = re_head * im_relation + im_head * re_relation
            score = re_score * re_tail + im_score * im_tail

        score = score.sum(dim = 2)
        return score

    def RotatE(self, head, relation, tail, mode):
        re_head, im_head = torch.chunk(head, 2, dim=2)
        re_tail, im_tail = torch.chunk(tail, 2, dim=2)

        phase_relation = relation / (self.embedding_range.item()/math.pi)

        re_relation = torch.cos(phase_relation)
        im_relation = torch.sin(phase_relation)

        if mode == 'head-batch':
            re_score = re_relation * re_tail + im_relation * im_tail
            im_score = re_relation * im_tail - im_relation * re_tail
            re_score = re_score - re_head
            im_score = im_score - im_head
        else:
            re_score = re_head * re_relation - im_head * im_relation
            im_score = re_head * im_relation + im_head * re_relation
            re_score = re_score - re_tail
            im_score = im_score - im_tail

        score = torch.stack([re_score, im_score], dim = 0)
        score = score.norm(dim = 0)

        score = self.gamma.item() - score.sum(dim=2)
        return score

    @staticmethod
    def train_step(model, optimizer, train_iterator, args):
        optimizer.zero_grad()

        positive_sample, negative_sample, subsampling_weight, mode, true = next(train_iterator)
        if args['cuda']:
            positive_sample = positive_sample.cuda()
            negative_sample = negative_sample.cuda()
            subsampling_weight = subsampling_weight.cuda()

        if mode == 'neg':
            negative_score = torch.zeros(positive_sample.shape)
            if args['cuda']:
                negative_score = negative_score.cuda()
        else:
            negative_score, _ = model((positive_sample, negative_sample), mode=mode)
            negative_score = F.logsigmoid(-negative_score).mean(dim=1)

        fwd_mode = 'neg' if mode == 'neg' else 'single'
        positive_score, attr_loss = model(positive_sample, mode=fwd_mode)
        positive_score = F.logsigmoid(positive_score).squeeze(dim=1)

        if mode != 'neg':
            positive_sample_loss = -(subsampling_weight * positive_score).sum() / subsampling_weight.sum()
            negative_sample_loss = -(subsampling_weight * negative_score).sum() / subsampling_weight.sum()
            loss = (positive_sample_loss + negative_sample_loss) / 2
        else:
            positive_sample_loss = -positive_score.sum()
            negative_sample_loss = torch.tensor([[0.]])
            loss = positive_sample_loss
        
        loss += attr_loss

        loss.backward()
        optimizer.step()

        stats = {
            'pos_loss': round(positive_sample_loss.item(), 6),
            'neg_loss': round(negative_sample_loss.item(), 6),
            'loss': round(loss.item(), 6),
            'attr_loss': round(float(attr_loss), 4)
        }
        return stats



import context

from zincbase import KB

kb = KB()
kb.seed(555)

kb.store('person(tom)')
kb.store('person(shamala)')
kb.store('knows(tom, shamala)')
assert kb.neighbors('tom') == [('shamala', [{'pred': 'knows'}])]

kb.node('tom')['grains'] = 0

tom = kb.node('tom')
assert tom.grains == 0
assert tom.i_dont_exist is None
assert tom['i_dont_exist'] is None

kb.node('shamala').grains = 4
shamala = kb.node('shamala')
assert 'grains' in shamala
assert 'grains' in shamala.attrs
assert shamala.grains == 4
shamala.grains += 1
assert shamala.grains == 5
assert shamala['grains'] == 5
shamala['grains'] += 1
assert shamala['grains'] == 6

kb.store('person(jeraca)')
kb.node('jeraca').grains = 3

zero_grains = list(kb.filter(lambda x: x['grains'] == 0))
assert len(zero_grains) == 1
assert zero_grains[0] == 'tom'
assert zero_grains[0] != 'shamala'

zero_anything = list(kb.filter(lambda x: x['anything'] == 0))
assert len(zero_anything) == 0

more_grains = kb.filter(lambda x: x['grains'] >= 3)
assert next(more_grains) in ['shamala', 'jeraca']
assert next(more_grains) in ['shamala', 'jeraca']

more_grains = kb.filter(lambda x: x['grains'] >= 3, candidate_nodes=['shamala'])
as_list = list(more_grains)
assert as_list == ['shamala']

more_grains = kb.filter(lambda x: x['grains'] >= 3, candidate_nodes=[])
as_list = list(more_grains)
assert as_list == []

some_or_no_grains = kb.filter(lambda x: x['grains'] >= -1, candidate_nodes=['tom', 'shamala'])
as_list = list(some_or_no_grains)
assert len(as_list) == 2
assert as_list[0] in ['tom', 'shamala']
assert as_list[1] in ['tom', 'shamala']
assert as_list[0] != as_list[1]

nodes = kb.filter(lambda x: True)
as_list = list(nodes)
assert len(as_list) == 3
jeraca = kb.node('jeraca')
assert len(jeraca.neighbors) == 0
shamala = kb.node('shamala')
assert len(shamala.neighbors) == 0
tom = kb.node('tom')
assert len(tom.neighbors) == 1
assert tom.neighbors[0][0] == 'shamala'
assert len(tom.neighbors[0][1]) == 1
assert tom.neighbors[0][1][0]['pred'] == 'knows'

fn_was_called = False
def watch_fn(node, prev_val):
    global fn_was_called
    fn_was_called = True
    assert prev_val == 0
    assert node.grains == 1
    assert len(node.neighbors) == 1
    assert kb.node(node.neighbors[0][0]) == 'shamala'

nights_watch = tom.watch('grains', watch_fn)
tom.grains += 1
assert fn_was_called
fn_was_called = False
tom.remove_watch(nights_watch)
tom.grains += 1
assert not fn_was_called
nights_watch = tom.watch('grains', watch_fn)
tom.remove_watch('grains')
tom.grains += 1
assert not fn_was_called

kb.store('node(i_am_node)', node_attributes=[{'foo': 'bar'}])
new_node = kb.node('i_am_node')
assert new_node.foo == 'bar'
new_node.foo = 'baz'
new_node = kb.node('i_am_node')
assert new_node.foo == 'baz'
kb.store('connected_nodes(3, 4)', node_attributes=[{'x': 3}, {'x': 4}], edge_attributes={'power_level': 3})
_3 = kb.node(3)
_4 = kb.node(4)
assert _3.x == 3
assert _4.x == 4
assert kb.edge(3, 'connected_nodes', 4).power_level == 3
kb.edge(3, 'connected_nodes', 4).power_level = 'high'
assert kb.edge(3, 'connected_nodes', 4).power_level == 'high'

print('All attribute tests passed.')


# Neural Network tests

# Possible that tests could flake given stochasticity of NN
# but they are fairly relaxed.

import context
from zincbase import KB

kb = KB()
kb.seed(555)

people = ['john', 'oleg', 'tom', 'vedant', 'akshay', 'todd']
for person in people:
    kb.store('works_at({}, primer)'.format(person))
other_people = ['other1', 'other2', 'other3', 'other4', 'other5', 'other6']
for person in other_people:
    kb.store('works_at({}, zillow)'.format(person))
kb.store('based_in(primer, bay_area)')
kb.store('based_in(zillow, seattle)')

for person in people:
    kb.store('lives_in({}, bay_area)'.format(person))

for person in other_people:
    kb.store('lives_in({}, seattle)'.format(person))

kb.store('knows(tom, john)'); kb.store('knows(tom, oleg)'); kb.store('knows(akshay, john)'); kb.store('knows(tom, todd)'); kb.store('knows(vedant, akshay)')
kb.store('knows(other2, other1)'); kb.store('knows(other6, other5)'); kb.store('knows(other1, other2)'); kb.store('knows(other4, other3)'); kb.store('knows(other3, other2)')
kb.store('knows(tom, other4)')
kb.store('lives_in(shamala, bay_area)'); kb.store('lives_in(mary, seattle)')
kb.store('associated_with(zillow, amazon)'); kb.store('associated_with(primer, google)')
kb.store('associated_with(amazon, zillow)'); kb.store('associated_with(google, primer)')
kb.store('associated_with(amazon, microsoft)'); kb.store('associated_with(google, facebook)')
kb.store('based_in(microsoft, seattle)'); kb.store('based_in(facebook, bay_area)')

kb.build_kg_model(cuda=False, embedding_size=30)
# Ideally use bs=1 to overfit on this small dataset
# bs=2 at least checks that it works with > 1 bs
kb.train_kg_model(steps=8001, batch_size=1)

# # # # # # # # # # # # # # # # # # # # # # # #
# Link prediction test
# # # # # # # # # # # # # # # # # # # # # # # #
bay_prob = kb.estimate_triple_prob('tom', 'lives_in', 'bay_area')
sea_prob = kb.estimate_triple_prob('tom', 'lives_in', 'seattle')
assert bay_prob > 2 * sea_prob

bay_prob = kb.estimate_triple_prob('shamala', 'lives_in', 'bay_area')
sea_prob = kb.estimate_triple_prob('shamala', 'lives_in', 'seattle')
assert bay_prob > 2 * sea_prob

sea_prob = kb.estimate_triple_prob('other1', 'lives_in', 'seattle')
bay_prob = kb.estimate_triple_prob('other1', 'lives_in', 'bay_area')
assert sea_prob > 2 * bay_prob
sea_prob = kb.estimate_triple_prob('other2', 'lives_in', 'seattle')
bay_prob = kb.estimate_triple_prob('other2', 'lives_in', 'bay_area')
assert sea_prob > 2 * bay_prob

print('All basic NN tests passed.')


"""Test negative examples using Countries.

The main idea here is that if we explicitly enter some false facts (signalling
to the KB that they are false, it should make less-wrong predictions
for them, versus just going by its own synthetic negative examples.)

It may have the side effect of pushing UP the probability of other wrong
triples, see e.g. "canada in asia" below.
"""
import context

from zincbase import KB

kb = KB()
kb.seed(555)

kb.from_csv('./assets/countries_s1_train.csv', delimiter='\t')

rule_num = kb.store('~locatedin(canada, africa)')

b = list(kb.query('locatedin(canada, X)'))
assert len(b) == 1; assert b[0]['X'] == 'northern_america'
assert kb.delete_rule(rule_num)

kb.build_kg_model(cuda=False, embedding_size=100)

kb.train_kg_model(steps=500, batch_size=512, neg_ratio=0.01)

canada_in_africa_naive = kb.estimate_triple_prob('canada', 'locatedin', 'africa')
canada_in_asia_naive = kb.estimate_triple_prob('canada', 'locatedin', 'asia')

austria_neighbors_spain_naive = kb.estimate_triple_prob('austria', 'neighbor', 'spain')
austria_neighbors_france_naive = kb.estimate_triple_prob('austria', 'neighbor', 'france')

kb = KB()
kb.seed(555)
kb.from_csv('./assets/countries_s1_train.csv', delimiter='\t')
kb.store('~locatedin(canada, africa)')
kb.store('~neighbor(austria, spain)')

kb.build_kg_model(cuda=False, embedding_size=100)
kb.train_kg_model(steps=500, batch_size=512, neg_ratio=0.1)

canada_in_africa_explicit = kb.estimate_triple_prob('canada', 'locatedin', 'africa')
canada_in_asia_explicit = kb.estimate_triple_prob('canada', 'locatedin', 'asia')
austria_neighbors_spain_explicit = kb.estimate_triple_prob('austria', 'neighbor', 'spain')
austria_neighbors_france_explicit = kb.estimate_triple_prob('austria', 'neighbor', 'france')

assert canada_in_africa_naive > canada_in_africa_explicit
assert austria_neighbors_spain_naive > austria_neighbors_spain_explicit

print('All negative example tests passed.')


def test_index(server_and_args):
    app, server, args = server_and_args
    response = app.test_client().get('/')
    assert response.status_code == 200
    data = response.data.decode('utf-8')
    assert '<title>Zincbase Graph Server</title>' in data
    assert 'src="bundle.js"' in data


"""Calculate the Area-Under-the-Curve Receiver Operating Characteristic

A funny measure that combines precision and recall. Sklearn can't agree how
to implement it for multiclass; this version is from fbrundu on
https://github.com/scikit-learn/scikit-learn/issues/3298
"""

from sklearn.metrics import roc_auc_score
from sklearn.preprocessing import LabelBinarizer

def calc_auc_roc(truth, pred, average="macro"):

    lb = LabelBinarizer()
    lb.fit(truth)
    truth = lb.transform(truth)
    pred = lb.transform(pred)
    return roc_auc_score(truth, pred, average=average)



class GraphCaster:
    def __init__(self, redis_address='redis://'):
        """Create a graph cast, so that a KB can be displayed
        on localhost:5000. It won't work if you installed basic
        Zincbase -- try `pip install zincbase[web]`.
        :param str redis_address: URL of the redis instance
        the graph cast should use. For local use, let the
        default stand, provided you have `docker run -p 6379:6379 -d redis`
        """
        try:
            from flask_socketio import SocketIO, emit
        except ImportError:
            print('Please install zincbase[web]')
            return False
        self.redis = redis_address
        self.socketio = SocketIO(message_queue=self.redis)
        self.socketio.emit('reset')
        self.node_update_queue = []
        self.edge_update_queue = []
   
    def add_node(self, node):
        """Add a node to the graph cast.

        :param Node node: A node in the KB.
        """
        attrs = { 'id': str(node) }
        attrs.update(node.attrs)
        self.socketio.emit('addNode', attrs, json=True)
    
    def add_edge(self, from_node, to_node, attributes):
        """Add an edge to the graph cast.
        """
        attrs = { 'source': str(from_node), 'target': str(to_node) }
        attrs.update(attributes)
        self.socketio.emit('addLink', attrs, json=True)
    
    def update_node(self, node, defer=False):
        """Update a node in the graph cast with its current attributes in the KB.
        
        :param Node node: The node to update
        :param bool defer: If False, send immediately (and cause immediate
        re-rendering on the client.) If True, batch and wait until `batch_update()`
        to send updates together and re-render only once.
        """
        attrs = { 'id': str(node) }
        attrs.update({ 'attributes': node.attrs })
        if not defer:
            self.socketio.emit('updateNode', attrs, json=True)
        else:
            self.node_update_queue.append(attrs)
    
    def update_edge(self, edge, defer=False):
        """Update an edge in the graph cast with its current attributes in the KB.
        
        :param str sub, pred, obj: The edge, described as subject, predicate, object
        :param KB kb: The knowledge base in which the edge exists
        :param bool defer: If False, send immediately (and cause immediate
        re-rendering on the client.) If True, batch and wait until `batch_update()`
        to send updates together and re-render only once.
        """
        attrs = { 'attributes': edge.attrs }
        attrs.update({'from': edge._sub, 'pred': edge._pred, 'to': edge._ob})
        if not defer:
            self.socketio.emit('updateEdge', attrs, json=True)
        else:
            self.edge_update_queue.append(attrs)
    
    def batch_update(self):
        """Perform a batch update. Any `update_node` or `update_edge` calls that
        were made with `defer=True` will now be sent to the frontend.
        """
        self.socketio.emit('batchUpdateNode', self.node_update_queue, json=True)
        self.socketio.emit('batchUpdateEdge', self.edge_update_queue, json=True)
        self.node_update_queue = []
        self.edge_update_queue = []

    def reset(self):
        """If any web client was already listening, reset it"""
        self.socketio.emit('reset')

    def render(self, node_color=0x11bb88, node_size=10, node_opacity=0.9,
               node_label='id', node_visibility=True, edge_label='pred',
               edge_opacity=1, edge_color=0x333333, edge_size=0,
               edge_visibility=True, arrow_size=0, arrow_color=0x000001,
               label_node=False, label_node_color='black', label_node_height=3,
               label_node_offset=1, label_edge=False, label_edge_color='black',
               label_edge_height=3, label_edge_offset=1,
               bg_color=0xffffff, engine='d3'):
        """Perform the initial setup/rendering of the current graph.

        :param node_color: Either a 24bit RGB int (such as 0xFF001A) or a string
        containing a Javascript function which takes `node` as an argument, for
        example `node => node.color`
        :param node_size: Either a number >= 0 or a string containing a Javascript
        function, for example `node => Math.log(node.enormity)`
        :param node_label: Either a string representing a property of the node
        to display (on hover) as its label, or a Javascript function returning a string.
        All nodes have a property called `id` which is their name/string repr.
        :param node_visibility: Either a string representing a property of the node
        which evaluates truthy/falsy (in Javascript) to determine whether to display
        the node, or a JS function that returns true or false, or True/False.
        :param label_node: If True, nodes will be labeled with `node_label`. Unlike
        `node_label`, which only displays on hover, this is a permanent text. Note
        that the value updates when the value of `node[node_label]` changes (in Python).
        :param label_node_color: RGB value for the color of a node's permanent label
        :param label_node_height: Text height for the node's permanent label
        :param label_node_offset: Integer specifying how far out from the node the 
        label should appear. Default is 1 unit on the z-axis.
        :param edge_visibility: Either a string representing a property of the edge
        which evaluates truthy/falsy (in Javascript) to determine whether to display
        the edge, or a JS function that returns true or false, or True/False.
        :param edge_label: Either a string representing a property of an edge
        to display (on hover) as its label, or a Javascript function returning a string.
        Defaults to the predicate.
        :param float edge_opacity: Opacity of the edges, from 0-1
        :param edge_color: Either a 24bit RGB int or a string containing a Javascript
        function which takes `edge` as an argument, for example `edge => edge.color`.
        :param edge_size: The width of an edge. Either a number >= 0 (where 0 means 1px)
        or a string containing a Javascript function.
        :param label_edge: If True, nodes will be labeled with `edge_label`. Unlike
        `edge_label`, which only displays on hover, this is a permanent text. Note
        that the value updates when the value of `edge[edge_label]` changes (in Python).
        :param label_edge_color: RGB value for the color of a edge's permanent label
        :param label_edge_height: Text height for the edge's permanent label
        :param label_edge_offset: Integer specifying how far out from the edge the 
        label should appear. Default is 1 unit on the z-axis.
        :param int arrow_size: If >0, display directional arrows on edges of that size.
        :param int arrow_color: Color of arrows (if arrow_size > 0)
        :param int bg_color: Hex background color for the graph, e.g. 0xFF0000 is red.
        :param str engine: Specify d3 or ngraph. ngraph is faster but can be buggy, and
        is only really suitable for static graphs. The layouts can look different also.
        """
        if label_node:
            label_node = {
                'color': 'black',
                'height': 3,
                'offset': node_size + label_node_offset
            }
        if label_edge:
            label_edge = {
                'color': 'black',
                'height': 3,
                'offset': edge_size + label_edge_offset
            }
        attributes = { 'node_color': node_color, 'node_size': node_size,
                       'node_opacity': node_opacity, 'node_label': node_label,
                       'node_visibility': node_visibility, 'edge_visibility': edge_visibility,
                       'edge_opacity': edge_opacity, 'edge_color': edge_color,
                       'edge_size': edge_size, 'edge_label': edge_label,
                       'arrow_size': arrow_size, 'arrow_color': arrow_color,
                       'label_node': label_node, 'label_edge': label_edge,
                       'engine': engine, 'bg_color': bg_color }
        self.socketio.emit('render', attributes, json=True)

    def from_kb(self, kb):
        # TODO Permit this to work with subsampling the KB
        for node in kb.G.nodes:
            self.add_node(kb.node(node))
        for from_node, to_node, edge_attributes in kb.G.edges.data():
            self.add_edge(from_node=from_node, to_node=to_node, attributes=edge_attributes)


from pip._internal.req import parse_requirements
from setuptools import setup, find_packages

from zincbase import __version__

def _parse(filename):
    """Parse a requirements file, including `-r requirements.txt` references"""
    return [str(r.req) for r in parse_requirements(filename, session=False)]

setup(name='zincbase',
      version=__version__,
      description='A state of the art knowledge base',
      long_description=open("README.md", "r").read(),
      long_description_content_type="text/markdown",
      url='https://github.com/complexdb/zincbase',
      author='ComplexDB',
      author_email='tom@complexdb.com',
      license='MIT',
      packages=find_packages(exclude=['test']),
      classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
      ],
      include_package_data=True,
      install_requires=_parse("requirements.txt"),
      extras_require={
        'web': ['flask-socketio', 'redis', 'eventlet']
      },
      python_requires='>=3.6'
      )


import argparse

from .socket_server import serve

parser = argparse.ArgumentParser(prog='python3 -m zincbase.web')
parser.add_argument('--redis', help='Address of Redis instance (default: localhost:6379)',
                      default='localhost:6379', type=str)

if __name__ == '__main__':
    args = parser.parse_args()
    serve(args)


from collections import defaultdict
import copy

import networkx as nx

class Node:
    """Class representing a node in the KB.
    """
    def __init__(self, kb, name, data, watches=[]):
        super().__setattr__('_kb', kb)
        super().__setattr__('_name', name)
        super().__setattr__('_recursion_depth', 0)
        nx.set_node_attributes(self._kb.G, {self._name: data})
        self._watches = defaultdict(list)
        for watch in watches:
            self._watches[watch[0]].append(watch[1])
    
    def __repr__(self):
        return self._name
    
    def __eq__(self, comparator):
        return self._name == str(comparator)
    
    def __ne__(self, comparator):
        return self._name != str(comparator)
    
    def __iter__(self):
        for attr in self.attrs:
            yield(attr)

    def __getattr__(self, key):
        try:
            return self._kb.G.nodes(data=True)[self._name][key]
        except KeyError as e:
            return None

    def __setattr__(self, key, value):
        if self._kb._global_propagations > self._kb._PROPAGATION_LIMIT:
            return False
        if self._recursion_depth > self._kb._MAX_RECURSION:
            return False
        self._kb._global_propagations += 1
        super().__setattr__('_recursion_depth', self._recursion_depth + 1)
        attrs = self._kb.G.nodes(data=True)[self._name]
        prev_val = attrs.get(key, None)
        attrs.update({key: value})
        nx.set_node_attributes(self._kb.G, {self._name: attrs})
        if not self._kb._dont_propagate:
            for watch_fn in self._watches.get(key, []):
                watch_fn(self, prev_val)
        super().__setattr__('_recursion_depth', self._recursion_depth - 1)
        self._kb._global_propagations -= 1

    def __getitem__(self, key):
        return self.__getattr__(key)
    
    def __setitem__(self, key, value):
        return self.__setattr__(key, value)
    
    def __delitem__(self, key):
        del self._kb.G.nodes[self._name][key]
    
    @property
    def attrs(self):
        """Returns attributes of the node stored in the KB
        """
        attributes = self._kb.G.nodes(data=True)[self._name]
        attributes = copy.deepcopy(attributes)
        try:
            del attributes['_watches']
            del attributes['_new_neighbor_fn']
        except:
            pass
        return attributes
    
    @property
    def neighbors(self):
        """Returns the node's neighbors, in the format of tuples:
        [(neighbor_name, [{'pred': predicate aka edge_relation}])]
        """
        return self._kb.neighbors(self._name)
    
    def watch(self, attribute, fn):
        """Execute user-defined function when the value of attribute changes.
        Function takes two args: `node` which has access to all
        its own attributes, including neighbors and edges, and the second
        arg is the previous value of the attribute that changed.

        As cycles are possible in the graph, changes to a node attribute, that
        change its neighbors attributes etc, may eventually propagate back
        to change the original node's attribute again, ad infinitum until
        the stack explodes. To prevent this, in one "update cycle", more
        than `kb._MAX_RECURSION` updates will be rejected.

        :returns int: id of the watch

        :Example:

        >>> kb.store('node(node1)')
        >>> node = kb.node('node1')
        >>> node.grains = 3
        >>> print(node.grains)
        3
        >>> node.watch('grains', lambda x: print('grains changed to ' + x.grains))
        ('grains', 0)
        >>> node.grains += 1
        grains changed to 4

        """
        self._watches[attribute].append(fn)
        return (attribute, len(self._watches) - 1)
    
    def remove_watch(self, attribute_or_watch_id):
        """Stop watching `attribute_or_watch_id`.
        If it is a string, delete all watches for that attribute.
        If it is a tuple of (attribute, watch_id): delete that specific watch.
        """
        if isinstance(attribute_or_watch_id, tuple):
            self._watches[attribute_or_watch_id[0]].pop(attribute_or_watch_id[1])
        else:
            self._watches[attribute_or_watch_id] = []
    
    def watch_for_new_neighbor(self, fn):
        """Execute `fn` when node receives a new neighbor."""
        self.__setattr__('_new_neighbor_fn', fn)


import context
from zincbase import KB

kb = KB()
kb.store('a(b,c)')
kb.node('b')['is_letter'] = 1.0
assert kb.node('b').attrs == {'is_letter': 1.0}
assert 'is_letter' in kb.node('b')
kb.node('b')['is_letter'] = 2.0
del kb.node('b')['is_letter']
assert 'is_letter' not in kb.node('b').attrs
kb.node('b')['is_letter'] = 2.0
assert kb.node('b').attrs == {'is_letter': 2.0}
kb.edge('b', 'a', 'c').both_alpha = 1.0
assert kb.edge('b', 'a', 'c').attrs == {'both_alpha': 1.0}
kb.edge('b', 'a', 'c')['both_alpha'] = 2.0
assert kb.edge('b', 'a', 'c').attrs == {'both_alpha': 2.0}
assert kb.to_triples() == [('b', 'a', 'c')]
triples = kb.to_triples(data=True)
assert triples == [('b', 'a', 'c', {'is_letter': 2.0}, {'both_alpha': 2.0}, {}, False)]
kb.node('c').is_letter = 0.9
triples = kb.to_triples(data=True)
assert triples == [('b', 'a', 'c', {'is_letter': 2.0}, {'both_alpha': 2.0}, {'is_letter': 0.9}, False)]
neg_rule_idx = kb.store('~a(b,c)')
triples = kb.to_triples(data=True)
assert triples == [('b', 'a', 'c', {'is_letter': 2.0}, {'both_alpha': 2.0}, {'is_letter': 0.9}, True)]
kb.delete_rule(neg_rule_idx)
triples = kb.to_triples(data=True)
assert triples == [('b', 'a', 'c', {'is_letter': 2.0}, {'both_alpha': 2.0}, {'is_letter': 0.9}, False)]
kb.edge('b', 'a', 'c').truthiness = -1
triples = kb.to_triples(data=True)
assert triples == [('b', 'a', 'c', {'is_letter': 2.0}, {'both_alpha': 2.0, 'truthiness': -1}, {'is_letter': 0.9}, True)]
del kb.edge('b', 'a', 'c')['truthiness']
triples = kb.to_triples(data=True)
assert triples == [('b', 'a', 'c', {'is_letter': 2.0}, {'both_alpha': 2.0}, {'is_letter': 0.9}, False)]
edge = kb.edge('b', 'a', 'c')
assert edge['both_alpha'] == 2.0
assert 'both_alpha' in edge.attrs
del edge['both_alpha']
assert 'both_alpha' not in edge.attrs
edge['both_alpha'] = 1.0
assert 'both_alpha' in edge.attrs
assert edge.attrs['both_alpha'] == 1.0

was_called = False
def nights_watch(edge, prev_val):
    global was_called
    assert prev_val == 1.0
    was_called = True
edge.watch('both_alpha', nights_watch)
edge['both_alpha'] = 2.0
assert was_called == True
assert edge.attrs['both_alpha'] == 2.0
assert edge['both_alpha'] == 2.0

kb.store('edge(a,b)')
edge = kb.edge('a', 'edge', 'b')
assert str(edge) == 'a___edge___b'
assert edge.pred == 'edge'
assert edge.nodes == ['a', 'b']
edge.resistance = 3
assert edge.resistance == 3
assert edge['resistance'] == 3
was_called = False
def test_good_stuff(edge, prev_val):
    global was_called
    assert prev_val == 3
    assert edge.resistance == 4
    was_called = True
edge.watch('resistance', test_good_stuff)
edge.resistance += 1
assert was_called
assert 'resistance' in edge.attrs
assert 'resistance' in edge
del edge['resistance']
assert 'resistance' not in edge.attrs
assert 'resistance' not in edge

try:
    kb.node('no_exist')
    assert False
except KeyError:
    assert True

try:
    kb.edge('hello', 'i', 'no_exist')
    assert False
except KeyError:
    assert True

print('All graph tests passed.')


import copy

class Goal:
    def __init__(self, rule, parent=None, bindings={}):
        self.rule = rule
        self.parent = parent
        self.bindings = copy.deepcopy(bindings)
        self.idx = 0